from collections.abc import Callable, Collection, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Any, Union

from fastapi import Response
from fastapi.params import Depends
from fastapi.routing import APIRoute
from starlette.routing import BaseRoute

from cadwyn.exceptions import LintingError

from .._utils import DATACLASS_SLOTS, Sentinel
from .common import _HiddenAttributeMixin

HTTP_METHODS = {"GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "HEAD"}


@dataclass(**DATACLASS_SLOTS)
class EndpointAttributesPayload:
    # FastAPI API routes also have "endpoint" and "dependency_overrides_provider" fields.
    # We do not use them because:
    #   1. "endpoint" must not change -- otherwise this versioning is doomed
    #   2. "dependency_overrides_provider" is taken from router's attributes
    #   3. "response_model" must not change for the same reason as endpoint
    # The following for the same reason as endpoint:
    # * response_model_include: SetIntStr | DictIntStrAny
    # * response_model_exclude: SetIntStr | DictIntStrAny
    # * response_model_by_alias: bool
    # * response_model_exclude_unset: bool
    # * response_model_exclude_defaults: bool
    # * response_model_exclude_none: bool
    path: str
    response_model: Any
    status_code: int
    tags: list[Union[str, Enum]]
    # Adding/removing dependencies between versions seems like a bad choice.
    # It makes the system overly complex. Instead, we allow people to
    # overwrite all dependencies of a route at once. Hence you always know exactly
    # which dependencies have been specified, no matter how many migrations you have.

    dependencies: Sequence[Depends]
    summary: str
    description: str
    response_description: str
    responses: dict[Union[int, str], dict[str, Any]]
    deprecated: bool
    methods: set[str]
    operation_id: str
    include_in_schema: bool
    response_class: type[Response]
    name: str
    callbacks: list[BaseRoute]
    openapi_extra: dict[str, Any]
    generate_unique_id_function: Callable[[APIRoute], str]


@dataclass(**DATACLASS_SLOTS)
class EndpointHadInstruction(_HiddenAttributeMixin):
    endpoint_path: str
    endpoint_methods: set[str]
    endpoint_func_name: Union[str, None]
    attributes: EndpointAttributesPayload


@dataclass(**DATACLASS_SLOTS)
class EndpointExistedInstruction(_HiddenAttributeMixin):
    endpoint_path: str
    endpoint_methods: set[str]
    endpoint_func_name: Union[str, None]


@dataclass(**DATACLASS_SLOTS)
class EndpointDidntExistInstruction(_HiddenAttributeMixin):
    endpoint_path: str
    endpoint_methods: set[str]
    endpoint_func_name: Union[str, None]


@dataclass(**DATACLASS_SLOTS)
class EndpointInstructionFactory:
    endpoint_path: str
    endpoint_methods: set[str]
    endpoint_func_name: Union[str, None]

    @property
    def didnt_exist(self) -> EndpointDidntExistInstruction:
        return EndpointDidntExistInstruction(
            is_hidden_from_changelog=False,
            endpoint_path=self.endpoint_path,
            endpoint_methods=self.endpoint_methods,
            endpoint_func_name=self.endpoint_func_name,
        )

    @property
    def existed(self) -> EndpointExistedInstruction:
        return EndpointExistedInstruction(
            is_hidden_from_changelog=False,
            endpoint_path=self.endpoint_path,
            endpoint_methods=self.endpoint_methods,
            endpoint_func_name=self.endpoint_func_name,
        )

    def had(
        self,
        *,
        path: str = Sentinel,
        response_model: Any = Sentinel,
        status_code: int = Sentinel,
        tags: list[Union[str, Enum]] = Sentinel,
        dependencies: Sequence[Depends] = Sentinel,
        summary: str = Sentinel,
        description: str = Sentinel,
        response_description: str = Sentinel,
        responses: dict[Union[int, str], dict[str, Any]] = Sentinel,
        deprecated: bool = Sentinel,
        methods: list[str] = Sentinel,
        operation_id: str = Sentinel,
        include_in_schema: bool = Sentinel,
        response_class: type[Response] = Sentinel,
        name: str = Sentinel,
        callbacks: list[BaseRoute] = Sentinel,
        openapi_extra: dict[str, Any] = Sentinel,
        generate_unique_id_function: Callable[[APIRoute], str] = Sentinel,
    ):
        return EndpointHadInstruction(
            is_hidden_from_changelog=False,
            endpoint_path=self.endpoint_path,
            endpoint_methods=self.endpoint_methods,
            endpoint_func_name=self.endpoint_func_name,
            attributes=EndpointAttributesPayload(
                path=path,
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                dependencies=dependencies,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                methods=set(methods) if methods is not Sentinel else Sentinel,
                operation_id=operation_id,
                include_in_schema=include_in_schema,
                response_class=response_class,
                name=name,
                callbacks=callbacks,
                openapi_extra=openapi_extra,
                generate_unique_id_function=generate_unique_id_function,
            ),
        )


def endpoint(path: str, methods: list[str], /, *, func_name: Union[str, None] = None) -> EndpointInstructionFactory:
    _validate_that_strings_are_valid_http_methods(methods)

    return EndpointInstructionFactory(path, set(methods), func_name)


def _validate_that_strings_are_valid_http_methods(methods: Collection[str]):
    invalid_methods = set(methods) - HTTP_METHODS
    if invalid_methods:
        invalid_methods = ", ".join(sorted(invalid_methods))
        raise LintingError(
            f"The following HTTP methods are not valid: {invalid_methods}. "
            "Please use valid HTTP methods such as GET, POST, PUT, PATCH, DELETE, OPTIONS, HEAD.",
        )


AlterEndpointSubInstruction = Union[EndpointDidntExistInstruction, EndpointExistedInstruction, EndpointHadInstruction]



from pathlib import Path
import re
import sys


def transform_string(input_string):
    # Regex to match the pattern and capture the section after #
    pattern = r'\./(concepts\.md)#([a-z0-9-]+)'

    # Replacement function to transform captured groups
    def replacement(match):
        filename, section = match.groups()
        # Replace .md with / and dashes with underscores in the section part
        return f"./{filename[:-3]}/{section.replace('-', '_')}.md"

    # Apply the regex substitution
    return re.sub(pattern, replacement, input_string)


Path(sys.argv[1]).write_text(transform_string(Path(sys.argv[1]).read_text()))



from typing import Annotated, Generic, TypeVar, Union

from fastapi.testclient import TestClient
from pydantic import BaseModel, Field
from starlette.status import HTTP_200_OK, HTTP_204_NO_CONTENT, HTTP_422_UNPROCESSABLE_CONTENT

from cadwyn import RequestInfo, convert_request_to_next_version_for
from cadwyn.route_generation import VersionedAPIRouter
from cadwyn.structure.schemas import schema
from cadwyn.structure.versions import VersionChange
from tests.conftest import CreateVersionedApp

BoolT = TypeVar("BoolT", bound=Union[bool, int])


class GenericSchema(BaseModel, Generic[BoolT]):
    foo: Annotated[bool, Field(title="Foo", strict=True)]
    bar: Annotated[BoolT, Field(title="Bar")]


class ParametrizedSchema(GenericSchema[bool]):
    pass


class GenericVersionChange(VersionChange):
    description = "Generic"
    instructions_to_migrate_to_previous_version = (
        schema(GenericSchema).field("foo").didnt_have("strict"),
        schema(GenericSchema).field("foo").had(type=BoolT),
        schema(GenericSchema).field("bar").had(description="A bar"),
        schema(GenericSchema).field("bar").didnt_have("title"),
    )

    @convert_request_to_next_version_for(GenericSchema)
    def to_bool(request: RequestInfo) -> None:
        request.body["foo"] = bool(request.body["foo"])


class ParametrizedVersionChange(VersionChange):
    description = "Parametrized"
    instructions_to_migrate_to_previous_version = (
        schema(ParametrizedSchema).field("bar").had(description="A bar"),
        schema(ParametrizedSchema).field("bar").didnt_have("title"),
    )


router = VersionedAPIRouter()


@router.post("/generic", status_code=HTTP_204_NO_CONTENT)
async def route_with_generic_schema(dep: GenericSchema) -> None:
    pass


@router.post("/parametrized", status_code=HTTP_204_NO_CONTENT)
async def route_with_parametrized_schema(dep: ParametrizedSchema) -> None:
    pass


def test__router_generation__using_generic_schema_in_body(
    create_versioned_app: CreateVersionedApp,
):
    app = create_versioned_app(
        GenericVersionChange,
        router=router,
    )

    unversioned_client = TestClient(app)
    client_2000 = TestClient(app, headers={app.router.api_version_parameter_name: "2000-01-01"})
    client_2001 = TestClient(app, headers={app.router.api_version_parameter_name: "2001-01-01"})

    assert client_2000.post("/generic", json={"foo": 1, "bar": False}).status_code == HTTP_204_NO_CONTENT

    assert client_2001.post("/generic", json={"foo": True, "bar": 0}).status_code == HTTP_204_NO_CONTENT
    assert client_2001.post("/generic", json={"foo": 1, "bar": False}).status_code == HTTP_422_UNPROCESSABLE_CONTENT

    assert unversioned_client.get("/openapi.json?version=2000-01-01").status_code == HTTP_200_OK
    assert unversioned_client.get("/openapi.json?version=2001-01-01").status_code == HTTP_200_OK


def test__router_generation__using_parametrized_schema_in_body(
    create_versioned_app: CreateVersionedApp,
):
    app = create_versioned_app(
        ParametrizedVersionChange,
        router=router,
    )

    unversioned_client = TestClient(app)
    client_2000 = TestClient(app, headers={app.router.api_version_parameter_name: "2000-01-01"})
    client_2001 = TestClient(app, headers={app.router.api_version_parameter_name: "2001-01-01"})

    assert client_2000.post("/parametrized", json={"foo": True, "bar": False}).status_code == HTTP_204_NO_CONTENT
    assert client_2001.post("/parametrized", json={"foo": True, "bar": False}).status_code == HTTP_204_NO_CONTENT

    assert unversioned_client.get("/openapi.json?version=2000-01-01").status_code == HTTP_200_OK
    assert unversioned_client.get("/openapi.json?version=2001-01-01").status_code == HTTP_200_OK



import ast
import inspect
import sys
from collections.abc import Callable
from enum import Enum, auto
from types import GenericAlias, LambdaType
from typing import (  # noqa: UP035
    Any,
    List,
    Union,
    cast,
    get_args,
    get_origin,
)

import annotated_types

from cadwyn._utils import PlainRepr, UnionType
from cadwyn.exceptions import InvalidGenerationInstructionError

_LambdaFunctionName = (lambda: None).__name__  # pragma: no branch
NoneType = type(None)


# A parent type of typing._GenericAlias
_BaseGenericAlias = cast("type", type(List[int])).mro()[1]  # noqa: UP006

# type(list[int]) and type(List[int]) are different which is why we have to do this.
# Please note that this problem is much wider than just lists which is why we use typing._BaseGenericAlias
# instead of typing._GenericAlias.
GenericAliasUnion = Union[GenericAlias, _BaseGenericAlias]
GenericAliasUnionArgs = get_args(GenericAliasUnion)
if sys.version_info >= (3, 14):  # pragma: no cover
    # In Python 3.14, Union types are no longer subtypes of _BaseGenericAlias
    GenericAliasUnionArgs = (*GenericAliasUnionArgs, type(Union[int, str]))


def get_fancy_repr(value: Any) -> Any:
    if isinstance(value, annotated_types.GroupedMetadata) and hasattr(type(value), "__dataclass_fields__"):
        return transform_grouped_metadata(value)
    if isinstance(value, (list, tuple, set, frozenset)):
        return transform_collection(value)
    if isinstance(value, dict):
        return transform_dict(value)
    if isinstance(value, UnionType):
        return transform_union(value)
    if isinstance(value, GenericAliasUnionArgs):
        return transform_generic_alias(value)
    if value is None or value is NoneType:
        return transform_none(value)
    if isinstance(value, type):
        return transform_type(value)
    if isinstance(value, Enum):
        return transform_enum(value)
    if isinstance(value, auto):  # pragma: no cover # it works but we no longer use auto
        return transform_auto(value)
    if isinstance(value, LambdaType) and _LambdaFunctionName == value.__name__:
        return transform_lambda(value)
    if inspect.isfunction(value):
        return transform_function(value)
    else:
        return transform_other(value)


def transform_grouped_metadata(value: "annotated_types.GroupedMetadata"):
    empty_obj = type(value)

    modified_fields = [
        (key, getattr(value, key))
        for key in value.__dataclass_fields__  # pyright: ignore[reportAttributeAccessIssue]
        if getattr(value, key) != getattr(empty_obj, key)
    ]

    return PlainRepr(
        value.__class__.__name__
        + "("
        + ", ".join(f"{PlainRepr(key)}={get_fancy_repr(v)}" for key, v in modified_fields)
        + ")",
    )


def transform_collection(value: Union[list, tuple, set, frozenset]) -> Any:
    return PlainRepr(value.__class__(map(get_fancy_repr, value)))


def transform_dict(value: dict) -> Any:
    return PlainRepr(
        value.__class__((get_fancy_repr(k), get_fancy_repr(v)) for k, v in value.items()),
    )


def transform_generic_alias(value: GenericAliasUnion) -> Any:
    return f"{get_fancy_repr(get_origin(value))}[{', '.join(get_fancy_repr(a) for a in get_args(value))}]"


def transform_none(_: type[None]) -> Any:
    return "None"


def transform_type(value: type) -> Any:
    return value.__name__


def transform_enum(value: Enum) -> Any:
    return PlainRepr(f"{value.__class__.__name__}.{value.name}")


def transform_auto(_: auto) -> Any:  # pragma: no cover # it works but we no longer use auto
    return PlainRepr("auto()")


def transform_union(value: UnionType) -> Any:
    return "typing.Union[" + (", ".join(get_fancy_repr(a) for a in get_args(value))) + "]"


def transform_lambda(value: LambdaType) -> Any:
    # We clean source because getsource() can return only a part of the expression which
    # on its own is not a valid expression such as: "\n  .had(default_factory=lambda: 91)"
    return _get_lambda_source_from_default_factory(inspect.getsource(value).strip(" \n\t."))


def transform_function(value: Callable) -> Any:
    return PlainRepr(value.__name__)


def transform_other(value: Any) -> Any:
    return PlainRepr(repr(value))


def _get_lambda_source_from_default_factory(source: str) -> str:
    found_lambdas: list[ast.Lambda] = [
        node.value
        for node in ast.walk(ast.parse(source))
        if isinstance(node, ast.keyword) and node.arg == "default_factory" and isinstance(node.value, ast.Lambda)
    ]

    if len(found_lambdas) == 1:
        return ast.unparse(found_lambdas[0])
    # These two errors are really hard to cover. Not sure if even possible, honestly :)
    elif len(found_lambdas) == 0:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            f"No lambda found in default_factory even though one was passed: {source}",
        )
    else:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            "More than one lambda found in default_factory. This is not supported.",
        )


def pop_docstring_from_cls_body(cls_body: list[ast.stmt]) -> list[ast.stmt]:
    if (
        len(cls_body) > 0
        and isinstance(cls_body[0], ast.Expr)
        and isinstance(cls_body[0].value, ast.Constant)
        and isinstance(cls_body[0].value.value, str)
    ):
        return [cls_body.pop(0)]
    else:
        return []



from typing import Callable

import pytest
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.routing import Match, NoMatchFound
from starlette.testclient import TestClient

from cadwyn import Cadwyn
from cadwyn.route_generation import VersionedAPIRouter
from cadwyn.structure.versions import HeadVersion, Version, VersionBundle
from tests._resources.app_for_testing_routing import mixed_hosts_app


def test__populate_routes():
    versioned_routes = [
        route for router in mixed_hosts_app.router.versioned_routers.values() for route in router.routes
    ]
    assert sorted(mixed_hosts_app.router.routes, key=id) == sorted(
        mixed_hosts_app.router.unversioned_routes + versioned_routes,
        key=id,
    )


def test__header_routing():
    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "2022-02-11"})

    response = client.get("/v1/users/tom/83")
    assert response.status_code == 200
    assert response.json() == {"users": [{"username": "tom", "page": 83}]}

    response = client.get("/v1/")
    # its fine, because "/v1/" is defined in the lower version
    assert response.status_code == 200

    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "2022-01-10"})

    response = client.get("/v1/users")
    assert response.status_code == 200
    assert response.text == "All users"

    response = client.get("/v1/")
    assert response.status_code == 200

    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "2022-03-12"})

    response = client.get("/v1/users")
    # its fine, because /users is defined in the lower version
    assert response.status_code == 200

    response = client.get("/v1/")
    assert response.status_code == 200

    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "2025-01-01"})

    response = client.get("/v1/users")
    # its fine, because /users is defined in the lower version
    assert response.status_code == 200


@pytest.mark.parametrize("version", ["2022-04-19", "2022-05-01", "2025-11-12"])
def test__host_routing__backward__ok(version: str):
    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": version})

    response = client.get("/v1/doggies/tom")
    assert response.status_code == 200
    assert response.json() == {"doggies": [{"dogname": "tom"}]}


def test__host_routing__lowest_version__404():
    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "1993-11-15"})

    response = client.get("/v1/doggies/tom")
    assert response.status_code == 404


def test__host_routing__non_http():
    assert mixed_hosts_app.routes[-1].matches({"type": "websocket", "path": "/v1/"}) == (Match.NONE, {})


def test__host_routing__non_date_api_version_header__not_valid_format():
    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "2025-40-01"})

    response = client.get("/v1/users")
    assert response.status_code == 422
    assert response.json() == {
        "detail": [
            {
                "type": "date_from_datetime_parsing",
                "loc": ["header", "x-api-version"],
                "msg": "Input should be a valid date or datetime, month value is outside expected range of 1-12",
                "input": "2025-40-01",
                "ctx": {"error": "month value is outside expected range of 1-12"},
            }
        ]
    }


def test__host_routing__partial_match__error():
    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "2022-02-11"})

    response = client.post("/v1/users/tom/83")
    assert response.status_code == 405


def test__url_path_for__not_enough_params__error():
    with pytest.raises(
        NoMatchFound,
        match='No route exists for name "api:users" and params "username".',
    ):
        mixed_hosts_app.url_path_for("api:users", username="tom")


def test__url_path_for__not_enough_params__error2():
    with pytest.raises(
        NoMatchFound,
        match='No route exists for name "api" and params "path, username".',
    ):
        mixed_hosts_app.url_path_for("api", path="hellow", username="tom")


def test__lifespan_async():
    startup_complete = False
    shutdown_complete = False

    async def hello_world(request: Request):
        return PlainTextResponse("hello, world")

    async def run_startup():
        nonlocal startup_complete
        startup_complete = True

    async def run_shutdown():
        nonlocal shutdown_complete
        shutdown_complete = True

    app = Cadwyn(
        versions=VersionBundle(Version("2022-11-16")),
        on_startup=[run_startup],
        on_shutdown=[run_shutdown],
    )
    app.add_route("/v1/", hello_world)

    assert not startup_complete
    assert not shutdown_complete
    with TestClient(app) as client:
        assert startup_complete
        assert not shutdown_complete
        client.get("/v1/")
    assert startup_complete
    assert shutdown_complete


def test__host_routing__partial_match__404():
    client = TestClient(mixed_hosts_app, headers={"X-API-VERSION": "1998-11-16"})

    response = client.get("/v1/doggies/tom")
    assert response.status_code == 200


async def get_default_version(req: Request):
    return "2023-04-14"


@pytest.mark.parametrize("default_version", ["2023-04-14", get_default_version])
def test__get_unversioned_endpoints__with_default_version(default_version: "str | Callable"):
    app = Cadwyn(
        versions=VersionBundle(HeadVersion(), Version("2023-04-14"), Version("2022-11-16")),
        api_version_default_value=default_version,
    )

    router = VersionedAPIRouter()

    @app.get("/my_duplicated_route")
    def get_my_unversioned_number():
        return 11

    @router.get("/my_duplicated_route")
    def get_my_versioned_number():
        return 83

    @router.get("/my_single_route")
    def get_my_versioned_number_2():
        return 52

    app.generate_and_include_versioned_routers(router)

    with TestClient(app) as client:
        resp = client.get("/docs")
        assert resp.status_code == 200, resp.json()

        resp = client.get("/docs?version=2023-04-14")
        assert resp.status_code == 200, resp.json()

        resp = client.get("/docs?version=2022-11-16")
        assert resp.status_code == 200, resp.json()

        resp = client.get("/my_duplicated_route")
        assert resp.status_code == 200, resp.json()
        assert resp.json() == 11

        resp = client.get("/my_duplicated_route", headers={"X-API-VERSION": "2023-04-14"})
        assert resp.status_code == 200, resp.json()
        assert resp.json() == 83

        resp = client.get("/my_single_route")
        assert resp.status_code == 200, resp.json()
        assert resp.json() == 52



# Note that this bug also affects 3.10+ but only on windows
# https://github.com/pyreadline/pyreadline/issues/65
import collections
from collections.abc import Callable

collections.Callable = Callable  # pyright: ignore[reportAttributeAccessIssue]



import uuid

from cadwyn import VersionedAPIRouter
from cadwyn.applications import Cadwyn
from cadwyn.structure import (
    RequestInfo,
    ResponseInfo,
    Version,
    VersionBundle,
    VersionChange,
    convert_request_to_next_version_for,
    convert_response_to_previous_version_for,
    endpoint,
    schema,
)
from cadwyn.structure.versions import HeadVersion
from pydantic import BaseModel, Field


class BaseUser(BaseModel):
    pass


class UserCreateRequest(BaseUser):
    default_address: str
    addresses_to_create: list[str] = Field(default_factory=list)


class UserResource(BaseUser):
    id: uuid.UUID


class UserAddressResource(BaseModel):
    id: uuid.UUID
    value: str


class UserAddressResourceList(BaseModel):
    data: list[UserAddressResource]


class ChangeAddressToList(VersionChange):
    description = "Change vat id to list"
    instructions_to_migrate_to_previous_version = (
        schema(BaseUser).field("addresses").didnt_exist,
        schema(BaseUser).field("address").existed_as(type=str, info=Field()),
    )

    @convert_request_to_next_version_for(UserCreateRequest)
    def change_address_to_multiple_items(request: RequestInfo) -> None:
        request.body["addresses"] = [request.body.pop("address")]

    @convert_response_to_previous_version_for(UserResource)
    def change_addresses_to_single_item(response: ResponseInfo) -> None:
        # Need to assert addresses length somewhere in business logic
        response.body["address"] = response.body["addresses"][0]


class ChangeAddressesToSubresource(VersionChange):
    description = "Change vat ids to subresource"
    instructions_to_migrate_to_previous_version = (
        schema(BaseUser)
        .field("addresses")
        .existed_as(type=list[str], info=Field()),
        schema(UserCreateRequest).field("default_address").didnt_exist,
        endpoint(
            "/{api_version}/users/{user_id}/addresses", ["GET"]
        ).didnt_exist,
    )

    @convert_request_to_next_version_for(UserCreateRequest)
    def change_addresses_to_default_address(request: RequestInfo):
        request.body["default_address"] = request.body["addresses"].pop(0)
        # Save data to still be able to keep the old behavior of creating addresses
        request.body["addresses_to_create"] = request.body.pop("addresses")

    @convert_response_to_previous_version_for(UserResource)
    def change_addresses_to_list(response: ResponseInfo) -> None:
        response.body["addresses"] = [
            id["value"] for id in response.body["_prefetched_addresses"]
        ]


class RemoveAddressesToCreateFromLatest(VersionChange):
    description = (
        "In order to support old versions, we gotta have `addresses_to_create` "
        "located in head schemas but we do not need this field in latest schemas."
    )
    instructions_to_migrate_to_previous_version = (
        schema(UserCreateRequest).field("addresses_to_create").didnt_exist,
    )


version_bundle = VersionBundle(
    HeadVersion(RemoveAddressesToCreateFromLatest),
    Version("v10", ChangeAddressesToSubresource),
    Version("v9", ChangeAddressToList),
    Version("v8"),
)


router = VersionedAPIRouter(tags=["Users"])
database_parody = {}


@router.post("/{api_version}/users", response_model=UserResource)
async def create_user(user: UserCreateRequest):
    id_ = uuid.uuid4()
    database_parody[id_] = {"id": id_}
    addresses = create_user_addresses(
        id_, [user.default_address, *user.addresses_to_create]
    )
    return database_parody[id_] | {"_prefetched_addresses": addresses}


@router.get("/{api_version}/users/{user_id}", response_model=UserResource)
async def get_user(user_id: uuid.UUID):
    return {
        "id": user_id,
        "_prefetched_addresses": (await get_user_addresses(user_id))["data"],
    }


def create_user_addresses(user_id: uuid.UUID, addresses: list[str]):
    database_parody[f"addr_{user_id}"] = [
        {"id": uuid.uuid4(), "value": address} for address in addresses
    ]
    return database_parody[f"addr_{user_id}"]


@router.get(
    "/{api_version}/users/{user_id}/addresses",
    response_model=UserAddressResourceList,
)
async def get_user_addresses(user_id: uuid.UUID):
    return {"data": database_parody[f"addr_{user_id}"]}


app = Cadwyn(
    versions=version_bundle,
    api_version_parameter_name="api_version",
    api_version_format="string",
    api_version_location="path",
)
app.generate_and_include_versioned_routers(router)



from typing import NoReturn

import pytest
from fastapi import APIRouter, HTTPException, Security
from fastapi.testclient import TestClient

from cadwyn.applications import Cadwyn
from cadwyn.structure.versions import Version, VersionBundle


class ScarySecurity:
    """It's IMPORTANT that we use an instance of a class instead of a function because it can be properly copied.

    It's an edge case.
    """

    async def __call__(self) -> NoReturn:
        raise HTTPException(status_code=401, detail="Unauthorized")


scary_security = ScarySecurity()


@pytest.fixture
def cadwyn_app():
    router = APIRouter()

    @router.get("/hello")
    async def world(user=Security(scary_security)):
        return {"hello": "world", "user": user}

    app = Cadwyn(versions=VersionBundle(Version("2023-11-16")))
    app.include_router(router)
    app.generate_and_include_versioned_routers(router)

    return app


@pytest.fixture
def user_client(cadwyn_app: Cadwyn):
    with TestClient(app=cadwyn_app, base_url="http://test") as ac:
        yield ac


@pytest.fixture
def _override_auth_dependency(cadwyn_app: Cadwyn) -> None:
    cadwyn_app.dependency_overrides[scary_security] = lambda: {"id": 123}


def test__no_dependency_overrides_with_unversioned_routes(user_client: TestClient):
    response = user_client.get("/hello")
    assert response.status_code == 401, response.json()


def test__no_dependency_overrides_with_versioned_routes(user_client: TestClient):
    response = user_client.get("/hello", headers={"x-api-version": "2023-11-16"})
    assert response.status_code == 401, response.json()


def test__dependency_overrides_with_unversioned_routes(user_client: TestClient, _override_auth_dependency: None):
    response = user_client.get("/hello")
    assert response.status_code == 200, response.json()


def test__dependency_overrides_with_versioned_routes(user_client: TestClient, _override_auth_dependency: None):
    response = user_client.get("/hello", headers={"x-api-version": "2023-11-16"})
    assert response.status_code == 200, response.json()



import bisect
from collections.abc import Sequence
from contextvars import ContextVar
from functools import cached_property
from logging import getLogger
from typing import Any, Union

from fastapi.routing import APIRouter
from starlette.datastructures import URL
from starlette.responses import RedirectResponse
from starlette.routing import BaseRoute, Match
from starlette.types import Receive, Scope, Send

from cadwyn._internal.context_vars import DEFAULT_API_VERSION_VAR
from cadwyn._utils import same_definition_as_in
from cadwyn.middleware import APIVersionFormat
from cadwyn.structure.common import VersionType

_logger = getLogger(__name__)


class _RootCadwynAPIRouter(APIRouter):
    def __init__(
        self,
        *args: Any,
        api_version_parameter_name: str,
        api_version_var: ContextVar[Union[str, None]],
        api_version_format: APIVersionFormat,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.versioned_routers: dict[VersionType, APIRouter] = {}
        self.api_version_parameter_name = api_version_parameter_name.lower()
        self.api_version_var = api_version_var
        self.unversioned_routes: list[BaseRoute] = []
        self.api_version_format = api_version_format

    async def _get_routes_from_closest_suitable_version(self, version: VersionType) -> list[BaseRoute]:
        """Pick the versioned routes for the given version in case we failed to pick a concrete version

        If the app has two versions: 2022-01-02 and 2022-01-05, and the request header
        is 2022-01-03, then the request will be routed to 2022-01-02 version as it the closest
        version, but lower than the request header.

        Exact match is always preferred over partial match and a request will never be
        matched to the higher versioned route.

        We implement routing like this because it is extremely convenient with microservice
        architecture. For example, imagine that you have two Cadwyn services: Payables and Receivables,
        each defining its own API versions. Payables service might contain 10 versions while receivables
        service might contain only 2 versions because it didn't need as many breaking changes.
        If a client requests a version that does not exist in receivables -- we will just waterfall
        to some earlier version, making receivables behavior consistent even if API keeps getting new versions.
        """
        if self.api_version_format == "date":
            index = bisect.bisect_left(self.versions, version)
            # That's when we try to get a version earlier than the earliest possible version
            if index == 0:
                return []
            picked_version = self.versions[index - 1]
            self.api_version_var.set(picked_version)
            # as bisect_left returns the index where to insert item x in list a, assuming a is sorted
            # we need to get the previous item and that will be a match
            return self.versioned_routers[picked_version].routes
        return []  # pragma: no cover # This should not be possible

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if "router" not in scope:  # pragma: no cover
            scope["router"] = self

        if scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)
            return
        version = self.api_version_var.get(None)
        default_version_that_was_picked = DEFAULT_API_VERSION_VAR.get(None)

        # if version is None, then it's an unversioned request and we need to use the unversioned routes
        # if there will be a value, we search for the most suitable version
        if not version:
            routes = self.unversioned_routes
        elif version in self.versioned_routers:
            routes = self.versioned_routers[version].routes
        else:
            routes = await self._get_routes_from_closest_suitable_version(version)
        if default_version_that_was_picked:
            # We add unversioned routes to versioned routes because otherwise unversioned routes
            # will be completely unavailable when a default version is passed. So routes such as
            # /docs will not be accessible at all.

            # We use this order because if versioned routes go first and there is a versioned route that is
            # the same as an unversioned route -- the unversioned one becomes impossible to match.
            routes = self.unversioned_routes + routes
        await self.process_request(scope=scope, receive=receive, send=send, routes=routes)

    @cached_property
    def versions(self):
        return sorted(self.versioned_routers.keys())

    @same_definition_as_in(APIRouter.add_api_route)
    def add_api_route(self, *args: Any, **kwargs: Any):
        super().add_api_route(*args, **kwargs)
        self.unversioned_routes.append(self.routes[-1])

    @same_definition_as_in(APIRouter.add_route)
    def add_route(self, *args: Any, **kwargs: Any):
        super().add_route(*args, **kwargs)
        self.unversioned_routes.append(self.routes[-1])

    @same_definition_as_in(APIRouter.add_api_websocket_route)
    def add_api_websocket_route(self, *args: Any, **kwargs: Any):  # pragma: no cover
        super().add_api_websocket_route(*args, **kwargs)
        self.unversioned_routes.append(self.routes[-1])

    @same_definition_as_in(APIRouter.add_websocket_route)
    def add_websocket_route(self, *args: Any, **kwargs: Any):  # pragma: no cover
        super().add_websocket_route(*args, **kwargs)
        self.unversioned_routes.append(self.routes[-1])

    async def process_request(self, scope: Scope, receive: Receive, send: Send, routes: Sequence[BaseRoute]) -> None:
        # It's a copy-paste from starlette.routing.Router
        # but in this version self.routes were replaced with routes from the function arguments

        partial = None
        partial_scope = {}
        for route in routes:
            # Determine if any route matches the incoming scope,
            # and hand over to the matching route if found.
            match, child_scope = route.matches(scope)
            if match == Match.FULL:
                scope.update(child_scope)
                await route.handle(scope, receive, send)
                return None
            if match == Match.PARTIAL and partial is None:
                partial = route
                partial_scope = child_scope

        if partial is not None:
            #  Handle partial matches. These are cases where an endpoint is
            # able to handle the request, but is not a preferred option.
            # We use this in particular to deal with "405 Method Not Allowed".
            scope.update(partial_scope)
            return await partial.handle(scope, receive, send)

        if scope["type"] == "http" and self.redirect_slashes and scope["path"] != "/":
            redirect_scope = dict(scope)
            if scope["path"].endswith("/"):
                redirect_scope["path"] = redirect_scope["path"].rstrip("/")
            else:
                redirect_scope["path"] = redirect_scope["path"] + "/"

            for route in routes:
                match, child_scope = route.matches(redirect_scope)
                if match != Match.NONE:
                    redirect_url = URL(scope=redirect_scope)
                    response = RedirectResponse(url=str(redirect_url))
                    await response(scope, receive, send)
                    return None

        return await self.default(scope, receive, send)



import uuid
from collections.abc import Sequence
from contextvars import ContextVar
from copy import deepcopy
from enum import Enum
from pathlib import Path
from typing import Any, Union

import pytest
from fastapi import APIRouter, FastAPI
from fastapi.testclient import TestClient
from pydantic import BaseModel
from pytest_fixture_classes import fixture_class

from cadwyn import Cadwyn, VersionBundle, VersionedAPIRouter
from cadwyn._utils import same_definition_as_in
from cadwyn.schema_generation import SchemaGenerator, generate_versioned_models
from cadwyn.structure import Version, VersionChange
from cadwyn.structure.endpoints import AlterEndpointSubInstruction
from cadwyn.structure.enums import AlterEnumSubInstruction
from cadwyn.structure.schemas import AlterSchemaSubInstruction, SchemaHadInstruction
from cadwyn.structure.versions import HeadVersion

CURRENT_DIR = Path(__file__).parent
Undefined = object()


@pytest.fixture
def api_version_var():
    api_version_var = ContextVar("api_version_var")
    api_version_var.set(None)
    return api_version_var


@pytest.fixture
def random_uuid():
    return uuid.uuid4()


class CadwynTestClient(TestClient):
    @same_definition_as_in(TestClient.__init__)
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.app: Cadwyn


class TestClientWithHardcodedAPIVersion(CadwynTestClient):
    def __init__(
        self,
        *args,
        api_version_var: Union[ContextVar, None] = None,
        api_version: Union[str, None, object] = Undefined,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.api_version_var = api_version_var
        self.api_version = api_version

    def request(self, *args, **kwargs):
        if self.api_version is not Undefined and self.api_version_var:
            self.api_version_var.set(self.api_version)
        return super().request(*args, **kwargs)


def client(
    router: APIRouter,
    api_version: Any = Undefined,
    api_version_var: Union[ContextVar[Union[str, None]], None] = None,
):
    app = FastAPI()
    app.include_router(router)

    return TestClientWithHardcodedAPIVersion(app, api_version=api_version, api_version_var=api_version_var)


@pytest.fixture
def router() -> VersionedAPIRouter:
    return VersionedAPIRouter()


@fixture_class(name="create_runtime_schemas")
class CreateRuntimeSchemas:
    def __call__(self, *version_changes: type[VersionChange]) -> dict[str, SchemaGenerator]:
        return generate_versioned_models(VersionBundle(*versions(*version_changes)))


@fixture_class(name="create_versioned_app")
class CreateVersionedApp:
    api_version_var: ContextVar[Union[str, None]]
    router: VersionedAPIRouter

    def __call__(
        self,
        *version_changes: type[VersionChange],
        head_version_changes: Sequence[type[VersionChange]] = (),
        router: Union[VersionedAPIRouter, None] = None,
    ) -> Cadwyn:
        router = router or self.router
        app = Cadwyn(
            versions=VersionBundle(
                HeadVersion(*head_version_changes),
                *versions(*version_changes),
                api_version_var=self.api_version_var,
            )
        )
        app.generate_and_include_versioned_routers(router)
        app._cadwyn_initialize()
        return app


def versions(*version_changes: type[VersionChange]) -> list[Version]:
    versions = [Version("2000-01-01")]
    for i, change in enumerate(version_changes):
        versions.append(Version(f"{2001 + i}-01-01", change))
    return list(reversed(versions))


@fixture_class(name="create_versioned_clients")
class CreateVersionedClients:
    create_versioned_app: CreateVersionedApp
    api_version_var: ContextVar[Union[str, None]]

    def __call__(
        self,
        *version_changes: type[VersionChange],
        head_version_changes: Sequence[type[VersionChange]] = (),
        router: Union[VersionedAPIRouter, None] = None,
    ) -> dict[str, CadwynTestClient]:
        app = self.create_versioned_app(*version_changes, head_version_changes=head_version_changes, router=router)
        return {
            version: CadwynTestClient(app, headers={app.router.api_version_parameter_name: version})
            for version in reversed(app.router.versioned_routers)
        }


def version_change(
    *instructions: Union[
        SchemaHadInstruction, AlterSchemaSubInstruction, AlterEndpointSubInstruction, AlterEnumSubInstruction
    ],
    **body_items: Any,
) -> type[VersionChange]:
    return type(VersionChange)(
        "MyVersionChange",  # pyright: ignore
        (VersionChange,),
        {
            "description": "",
            "instructions_to_migrate_to_previous_version": instructions,
            **body_items,
        },
    )


def serialize_enum(enum: type[Enum]) -> dict[str, Any]:
    return {member.name: member.value for member in enum}


def assert_models_are_equal(model1: type[BaseModel], model2: type[BaseModel]):
    model1_schema = serialize_schema(model1.__pydantic_core_schema__)
    model2_schema = serialize_schema(model2.__pydantic_core_schema__)
    assert model1_schema == model2_schema


def serialize_schema(schema: Any):
    schema = deepcopy(schema)
    if "cls" in schema:
        schema_to_modify = schema
    else:
        schema_to_modify = schema["schema"]
    del schema_to_modify["cls"]
    if "model_name" in schema_to_modify["schema"]:
        del schema_to_modify["schema"]["model_name"]
    if "schema" in schema_to_modify["schema"]:
        del schema_to_modify["schema"]["schema"]["model_name"]
    del schema_to_modify["config"]["title"]
    return serialize_object(schema)


def serialize_object(obj: Any):
    if isinstance(obj, dict):
        obj.pop("schema_ref", None)
        obj.pop("ref", None)
        return {k: v.__name__ if callable(v) else serialize_object(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        modified_list = []
        for v in obj:
            if callable(v):
                v = v.__name__
            modified_list.append(serialize_object(v))
        return modified_list
    else:
        return obj



import re
from enum import Enum, auto

import pytest

from cadwyn.exceptions import (
    InvalidGenerationInstructionError,
)
from cadwyn.structure import enum
from tests.conftest import CreateRuntimeSchemas, serialize_enum, version_change


class EmptyEnum(Enum):
    """My docstring"""


class EnumWithOneMember(Enum):
    foo = 83


class EnumWithTwoMembers(Enum):
    foo = 90
    bar = 12


class EnumWithOneMemberAndMethods(Enum):
    foo = 83

    def _hello(self):
        return self.hello_member  # pyright: ignore[reportAttributeAccessIssue]

    @classmethod
    def world(cls) -> int:
        return cls.world_member  # pyright: ignore[reportAttributeAccessIssue]


def test__enum_had__should_copy_docstring_too(create_runtime_schemas: CreateRuntimeSchemas):
    models = create_runtime_schemas(version_change(enum(EmptyEnum).had(b=auto())))
    assert models["2000-01-01"][EmptyEnum].__doc__ == "My docstring"
    assert models["2001-01-01"][EmptyEnum].__doc__ == "My docstring"


def test__enum_had__original_enum_is_empty(create_runtime_schemas: CreateRuntimeSchemas):
    models = create_runtime_schemas(version_change(enum(EmptyEnum).had(b=auto())))
    assert serialize_enum(models["2000-01-01"][EmptyEnum]) == {"b": 1}


def test__enum_had__original_enum_has_methods__all_methods_are_preserved(
    create_runtime_schemas: CreateRuntimeSchemas,
):
    models = create_runtime_schemas(
        version_change(enum(EnumWithOneMemberAndMethods).had(hello_member=10, world_member=20))
    )
    model = models["2000-01-01"][EnumWithOneMemberAndMethods]
    assert model.foo.value == 83
    assert model.hello_member.value == 10  # pyright: ignore[reportAttributeAccessIssue]
    assert model.world_member.value == 20  # pyright: ignore[reportAttributeAccessIssue]
    assert model.foo._hello().value == 10
    assert model.world().value == 20  # pyright: ignore[reportAttributeAccessIssue]


def test__enum_had__original_enum_is_nonempty(create_runtime_schemas: CreateRuntimeSchemas):
    models = create_runtime_schemas(version_change(enum(EnumWithOneMember).had(b=7)))
    assert serialize_enum(models["2000-01-01"][EnumWithOneMember]) == {"foo": 83, "b": 7}


def test__enum_didnt_have__original_enum_has_one_member(create_runtime_schemas: CreateRuntimeSchemas):
    models = create_runtime_schemas(version_change(enum(EnumWithOneMember).didnt_have("foo")))
    assert serialize_enum(models["2000-01-01"][EnumWithOneMember]) == {}


def test__enum_didnt_have__original_enum_has_two_members(create_runtime_schemas: CreateRuntimeSchemas):
    models = create_runtime_schemas(version_change(enum(EnumWithTwoMembers).didnt_have("foo")))
    assert serialize_enum(models["2000-01-01"][EnumWithTwoMembers]) == {"bar": 12}


def test__enum_had__original_schema_is_empty(create_runtime_schemas: CreateRuntimeSchemas):
    models = create_runtime_schemas(version_change(enum(EmptyEnum).had(b=7)))
    assert serialize_enum(models["2000-01-01"][EmptyEnum]) == {"b": 7}


def test__enum_had__same_name_as_other_value__error(
    create_runtime_schemas: CreateRuntimeSchemas,
):
    with pytest.raises(
        InvalidGenerationInstructionError,
        match=re.escape(
            'You tried to add a member "foo" to "EnumWithOneMember" in '
            '"MyVersionChange" but there is already a member with that name and value.',
        ),
    ):
        create_runtime_schemas(version_change(enum(EnumWithOneMember).had(foo=83)))


def test__enum_didnt_have__nonexisting_name__error(
    create_runtime_schemas: CreateRuntimeSchemas,
):
    with pytest.raises(
        InvalidGenerationInstructionError,
        match=re.escape(
            'You tried to delete a member "hoo" from "EnumWithOneMember" in '
            '"MyVersionChange" but it doesn\'t have such a member.',
        ),
    ):
        create_runtime_schemas(version_change(enum(EnumWithOneMember).didnt_have("hoo")))



import re
from contextvars import ContextVar
from typing import Any, Union, get_args

import pytest
from pydantic import BaseModel

from cadwyn.applications import Cadwyn
from cadwyn.exceptions import CadwynError, CadwynStructureError, LintingError
from cadwyn.structure import (
    Version,
    VersionBundle,
    VersionChange,
    VersionChangeWithSideEffects,
    convert_request_to_next_version_for,
    convert_response_to_previous_version_for,
    endpoint,
    schema,
)
from cadwyn.structure.schemas import FieldChanges, PossibleFieldAttributes
from cadwyn.structure.versions import HeadVersion


class DummySubClass2000_001(VersionChangeWithSideEffects):  # noqa: N801
    description = "dummy description"
    instructions_to_migrate_to_previous_version = []


class DummySubClass2000_002(VersionChangeWithSideEffects):  # noqa: N801
    description = "dummy description2"
    instructions_to_migrate_to_previous_version = []


class DummySubClass2001(VersionChangeWithSideEffects):
    description = "dummy description3"
    instructions_to_migrate_to_previous_version = []


class DummySubClass2002(VersionChangeWithSideEffects):
    description = "dummy description4"
    instructions_to_migrate_to_previous_version = []


@pytest.fixture
def dummy_sub_class_without_version():
    class DummySubClassWithoutVersion(VersionChangeWithSideEffects):
        description = "dummy description4"
        instructions_to_migrate_to_previous_version = []

    return DummySubClassWithoutVersion


@pytest.fixture
def versions(api_version_var: ContextVar[Union[str, None]]):
    try:
        yield VersionBundle(
            Version("2002-01-01", DummySubClass2002),
            Version("2001-01-01", DummySubClass2001),
            Version("2000-01-01", DummySubClass2000_001, DummySubClass2000_002),
            Version("1999-01-01"),
            api_version_var=api_version_var,
        )
    finally:
        DummySubClass2002._bound_version_bundle = None
        DummySubClass2001._bound_version_bundle = None
        DummySubClass2000_001._bound_version_bundle = None
        DummySubClass2000_002._bound_version_bundle = None


class TestVersionChange:
    def test__description__not_set__should_raise_error(self):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "Version change description is not set on 'DummySubClass' but is required.",
            ),
        ):

            class DummySubClass(VersionChange):
                instructions_to_migrate_to_previous_version = []

    def test__instructions_to_migrate_to_previous_version__not_set__should_raise_error(self):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "Attribute 'instructions_to_migrate_to_previous_version' "
                "is not set on 'DummySubClass' but is required.",
            ),
        ):

            class DummySubClass(VersionChange):
                description = "dummy description"

    def test__instructions_to_migrate_to_previous_version__not_a_sequence__should_raise_error(self):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "Attribute 'instructions_to_migrate_to_previous_version' must be a sequence in 'DummySubClass'.",
            ),
        ):

            class DummySubClass(VersionChange):
                description = "dummy description"
                instructions_to_migrate_to_previous_version = True  # pyright: ignore[reportAssignmentType]

    def test__instructions_to_migrate_to_previous_version__non_instruction_specified_in_list__should_raise_error(self):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "Instruction 'True' is not allowed. Please, use the correct instruction types",
            ),
        ):

            class DummySubClass(VersionChange):
                description = "dummy description"
                instructions_to_migrate_to_previous_version = [True]  # pyright: ignore[reportAssignmentType]

    def test__non_instruction_attribute_set__should_raise_error(self):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "Found: 'dummy_attribute' attribute of type '<class 'str'>' in 'DummySubClass'. "
                "Only migration instructions and schema properties are allowed in version change class body.",
            ),
        ):

            class DummySubClass(VersionChange):
                description = "dummy description"
                instructions_to_migrate_to_previous_version = []
                dummy_attribute = "dummy attribute"

    @pytest.mark.parametrize(
        "version_change_type",
        [VersionChange, VersionChangeWithSideEffects],
    )
    def test__init_subclass__incorrect_subclass_hierarchy__should_raise_error(
        self,
        version_change_type: type[VersionChange],
    ):
        class DummySubClass(version_change_type):
            description = "dummy description"
            instructions_to_migrate_to_previous_version = []

        with pytest.raises(
            TypeError,
            match=re.escape(
                "Can't subclass DummySubSubClass as it was never meant to be subclassed.",
            ),
        ):

            class DummySubSubClass(DummySubClass):
                pass

    def test__init__instantiation_attempt__should_raise_error(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "Can't instantiate DummySubClass as it was never meant to be instantiated.",
            ),
        ):

            class DummySubClass(VersionChange):
                description = "dummy description"
                instructions_to_migrate_to_previous_version = []

            DummySubClass()  # this will raise a TypeError


class TestVersionChangeWithSideEffects:
    def test__is_applied__api_version_var_is_none__everything_is_applied(
        self,
        versions: VersionBundle,
        api_version_var: ContextVar[Union[str, None]],
    ):
        api_version_var.set(None)
        assert DummySubClass2002.is_applied is True
        assert DummySubClass2001.is_applied is True
        assert DummySubClass2000_001.is_applied is True
        assert DummySubClass2000_002.is_applied is True

    def test__is_applied__api_version_var_is_later_than_latest__everything_is_applied(
        self,
        versions: VersionBundle,
        api_version_var: ContextVar[Union[str, None]],
    ):
        api_version_var.set("2003-01-01")
        assert DummySubClass2002.is_applied is True
        assert DummySubClass2001.is_applied is True
        assert DummySubClass2000_001.is_applied is True
        assert DummySubClass2000_002.is_applied is True

    def test__is_applied__api_version_var_is_before_latest__latest_is_inactive(
        self,
        versions: VersionBundle,
        api_version_var: ContextVar[Union[str, None]],
    ):
        api_version_var.set("2001-01-01")
        assert DummySubClass2002.is_applied is False
        assert DummySubClass2001.is_applied is True
        assert DummySubClass2000_001.is_applied is True
        assert DummySubClass2000_002.is_applied is True

    def test__is_applied__api_version_var_is_at_earliest__everything_is_inactive(
        self,
        versions: VersionBundle,
        api_version_var: ContextVar[Union[str, None]],
    ):
        api_version_var.set("1999-03-01")
        assert DummySubClass2002.is_applied is False
        assert DummySubClass2001.is_applied is False
        assert DummySubClass2000_001.is_applied is False
        assert DummySubClass2000_002.is_applied is False

    def test__is_applied__api_version_var_set_and_version_change_class_not_in_versions__should_raise_error(
        self,
        dummy_sub_class_without_version: type[VersionChangeWithSideEffects],
        api_version_var: ContextVar[Union[str, None]],
    ):
        api_version_var.set("1999-03-01")
        with pytest.raises(
            CadwynError,
            match=re.escape(
                "You tried to check whether 'DummySubClassWithoutVersion' "
                "is active but it was never bound to any version.",
            ),
        ):
            assert dummy_sub_class_without_version.is_applied

    def test__is_applied__api_version_var_unset_and_version_change_class_not_in_versions__should_raise_error(
        self,
        dummy_sub_class_without_version: type[VersionChangeWithSideEffects],
    ):
        with pytest.raises(
            CadwynError,
            match=re.escape(
                "You tried to check whether 'DummySubClassWithoutVersion' "
                "is active but it was never bound to any version.",
            ),
        ):
            assert dummy_sub_class_without_version.is_applied


class TestVersionBundle:
    def test__init__incorrectly_sorted_versions(self, api_version_var: ContextVar[Union[str, None]]):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "Versions are not sorted correctly. Please sort them in descending order.",
            ),
        ):
            Cadwyn(
                versions=VersionBundle(
                    Version("2000-01-01"),
                    Version("2001-01-01"),
                    api_version_var=api_version_var,
                )
            )

    def test__init__one_version_change_attached_to_two_version_bundles__should_raise_error(
        self,
        dummy_sub_class_without_version: type[VersionChangeWithSideEffects],
        api_version_var: ContextVar[Union[str, None]],
    ):
        VersionBundle(
            Version("2001-01-01", dummy_sub_class_without_version),
            Version("2000-01-01"),
            api_version_var=api_version_var,
        )
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "You tried to bind version change 'DummySubClassWithoutVersion' to two different versions."
                " It is prohibited.",
            ),
        ):
            VersionBundle(
                Version("2001-01-01", dummy_sub_class_without_version),
                Version("2000-01-01"),
                api_version_var=api_version_var,
            )

    def test__init__one_version_change_attached_to_two_versions__should_raise_error(
        self,
        dummy_sub_class_without_version: type[VersionChangeWithSideEffects],
        api_version_var: ContextVar[Union[str, None]],
    ):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "You tried to bind version change 'DummySubClassWithoutVersion' to two different versions. "
                "It is prohibited.",
            ),
        ):
            VersionBundle(
                Version("2002-01-01", dummy_sub_class_without_version),
                Version("2001-01-01", dummy_sub_class_without_version),
                Version("2000-01-01"),
                api_version_var=api_version_var,
            )

    def test__init__two_versions_with_the_same_value__should_raise_error(
        self,
        api_version_var: ContextVar[Union[str, None]],
    ):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                "You tried to define two versions with the same value in the same VersionBundle: '2000-01-01'.",
            ),
        ):
            VersionBundle(
                Version("2000-01-01"),
                Version("2000-01-01"),
                api_version_var=api_version_var,
            )

    def test__init__no_dated_versions_defined__should_raise_error(self, api_version_var: ContextVar[Union[str, None]]):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape("You must define at least one non-head version in a VersionBundle."),
        ):
            VersionBundle(HeadVersion(), api_version_var=api_version_var)

    def test__init__version_change_in_the_first_version__should_raise_error(
        self,
        dummy_sub_class_without_version: type[VersionChangeWithSideEffects],
        api_version_var: ContextVar[Union[str, None]],
    ):
        with pytest.raises(
            CadwynStructureError,
            match=re.escape(
                'The first version "2000-01-01" cannot have any version changes. '
                "Version changes are defined to migrate to/from a previous version "
                "so you cannot define one for the very first version.",
            ),
        ):
            VersionBundle(
                Version("2000-01-01", dummy_sub_class_without_version),
                api_version_var=api_version_var,
            )


class SomeSchema(BaseModel):
    pass


def test__convert_response_to_previous_version_for__with_incorrect_args__should_raise_error():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Method 'my_conversion_method' must have only 1 parameter: response",
        ),
    ):

        @convert_response_to_previous_version_for(SomeSchema)
        def my_conversion_method(cls: Any, payload: Any):  # pragma: no branch
            raise NotImplementedError


def test__convert_response_to_previous_version_for__with_no_args__should_raise_error():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Method 'my_conversion_method2' must have only 1 parameter: response",
        ),
    ):

        @convert_response_to_previous_version_for(SomeSchema)
        def my_conversion_method2():  # pragma: no branch
            raise NotImplementedError


def test__convert_request_to_next_version_for__with_incorrect_args__should_raise_error():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Method 'my_conversion_method' must have only 1 parameter: request",
        ),
    ):

        @convert_request_to_next_version_for(SomeSchema)
        def my_conversion_method(cls: Any, payload: Any):  # pragma: no branch
            raise NotImplementedError


def test__convert_request_to_next_version_for__with_no_args__should_raise_error():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Method 'my_conversion_method2' must have only 1 parameter: request",
        ),
    ):

        @convert_request_to_next_version_for(SomeSchema)
        def my_conversion_method2():  # pragma: no branch
            raise NotImplementedError


def test__schema_field_had_arguments_are_in_sync_with_schema_field_didnt_have_typehints():
    parameter_names_in_field_had = FieldChanges.__dataclass_fields__
    parameter_names_in_field_didnt_have = get_args(PossibleFieldAttributes)
    assert set(parameter_names_in_field_had) == set(parameter_names_in_field_didnt_have)


def test__endpoint_instruction_factory_interface__with_wrong_http_methods__should_raise_error():
    with pytest.raises(
        LintingError,
        match=re.escape(
            "The following HTTP methods are not valid: DEATH, STRAND. "
            "Please use valid HTTP methods such as GET, POST, PUT, PATCH, DELETE, OPTIONS, HEAD.",
        ),
    ):
        endpoint("/test", ["DEATH", "STRAND"])


def test__schema_validator_existed__non_validator_was_passed__should_raise_error():
    def fake_validator(cls, value):
        raise NotImplementedError

    with pytest.raises(CadwynStructureError, match=re.escape("The passed function must be a pydantic validator")):
        schema(BaseModel).validator(fake_validator).existed


def test__schema_validator_existed__non_function_was_passed__should_raise_error():
    with pytest.raises(CadwynStructureError, match=re.escape("The passed function must be a pydantic validator")):
        schema(BaseModel).validator(CadwynStructureError).existed


def test__function_as_validator__should_raise_error():
    with pytest.raises(CadwynStructureError, match=re.escape("The passed function must be a pydantic validator")):
        schema(BaseModel).validator(CadwynStructureError).existed



import pytest
from dirty_equals import IsUUID
from fastapi.testclient import TestClient

from docs_src.how_to.version_with_path_and_numbers_instead_of_headers_and_dates.block001 import (
    app,
)


@pytest.fixture
def testclient() -> TestClient:
    return TestClient(app)


def test__v8(testclient: TestClient):
    response = testclient.post(
        "/v8/users", json={"address": "123 Example St"}
    ).json()
    assert response == {
        "id": IsUUID(4),
        "address": "123 Example St",
    }
    assert testclient.get(f"/v8/users/{response['id']}").json() == {
        "id": response["id"],
        "address": "123 Example St",
    }


def test__v9(testclient: TestClient):
    response = testclient.post(
        "/v9/users", json={"addresses": ["124", "567"]}
    ).json()
    assert response == {
        "id": IsUUID(4),
        "addresses": ["124", "567"],
    }

    assert testclient.get(f"/v9/users/{response['id']}").json() == {
        "id": response["id"],
        "addresses": ["124", "567"],
    }


def test__v10(testclient: TestClient):
    response = testclient.post(
        "/v10/users", json={"default_address": "wowee"}
    ).json()

    assert response == {
        "id": IsUUID(4),
    }

    assert testclient.get(f"/v10/users/{response['id']}").json() == {
        "id": response["id"]
    }

    assert testclient.get(f"/v10/users/{response['id']}/addresses").json() == {
        "data": [
            {"id": IsUUID(4), "value": "wowee"},
        ],
    }


def test__using_versions_not_present_in_versionbundle__waterfalling_shouldnt_work(
    testclient: TestClient,
):
    response = testclient.post("/v7/users", json={"address": "123 Example St"})
    assert response.status_code == 404

    response = testclient.post("/v11/users", json={"address": "123 Example St"})
    assert response.status_code == 404



"""The test is kept in a separate file to avoid accidental importing of Literal"""

from __future__ import annotations

from typing import Annotated  # Note: _no_ import of Literal. This triggers the bug!

import fastapi
import pytest
from fastapi import Depends
from fastapi.testclient import TestClient
from pydantic import BaseModel, Field

import cadwyn
from cadwyn import current_dependency_solver
from cadwyn.applications import Cadwyn
from cadwyn.route_generation import VersionedAPIRouter
from cadwyn.structure.schemas import schema
from cadwyn.structure.versions import Version, VersionBundle, VersionChange


def test__missing_an_import_used_in_annotations_with_from_future_import_annotations():
    """Regression test for https://github.com/zmievsa/cadwyn/issues/324"""

    class MySchema(BaseModel):
        foo: str = Field(coerce_numbers_to_str=True)

    class MyVersionChange(VersionChange):
        description = "Hello"
        instructions_to_migrate_to_previous_version = (
            schema(MySchema).field("foo").had(type=int),
            schema(MySchema).field("foo").didnt_have("coerce_numbers_to_str"),
        )

    app = Cadwyn(versions=VersionBundle(Version("2001-01-01", MyVersionChange), Version("2000-01-01")))
    router = VersionedAPIRouter()

    async def dep_with_solver(
        dependency_solver: Annotated[Literal[fastapi, cadwyn], Depends(current_dependency_solver)],  # noqa: F821  # pyright: ignore[reportUndefinedVariable]
    ):
        pass

    @router.post(
        "/test",
        dependencies=[Depends(dep_with_solver)],
    )
    async def route_with_inner_schema_forwardref(dep: MySchema) -> MySchema:
        return dep

    app.generate_and_include_versioned_routers(router)
    client_2000 = TestClient(app, headers={app.router.api_version_parameter_name: "2000-01-01"})
    with pytest.raises(
        ImportError,
        match="You are likely missing an import from typing such as typing.Literal which causes RecursionError",
    ):
        assert client_2000.post("/test", json={"foo": 1}).json() == {"foo": 1}



import re
from typing import Annotated, Any, Generic, TypeVar, Union

import pytest
from pydantic import BaseModel, Field

from cadwyn import schema
from cadwyn.exceptions import InvalidGenerationInstructionError
from cadwyn.structure.schemas import PossibleFieldAttributes
from tests.conftest import (
    CreateRuntimeSchemas,
    assert_models_are_equal,
    version_change,
)
from tests.test_schema_generation.test_schema_field import assert_field_had_changes_apply

BoolT = TypeVar("BoolT", bound=Union[bool, int])


class GenericSchema(BaseModel, Generic[BoolT]):
    foo: BoolT
    bar: bool


class ParametrizedSchema(GenericSchema[bool]):
    pass


ATTR_MAP: tuple[tuple[str, str], list[tuple[PossibleFieldAttributes, Any]]] = (
    ("attr", "value"),
    [
        ("title", "Foo"),
        ("description", "A foo"),
        ("repr", False),
    ],
)

###########
# GENERIC #
###########


def test__schema_field_existed_as__field_is_typevar(create_runtime_schemas: CreateRuntimeSchemas):
    schemas = create_runtime_schemas(
        version_change(
            schema(GenericSchema)
            .field("baz")
            .existed_as(
                type=BoolT,
            ),
        ),
    )

    class ExpectedSchema(BaseModel):
        foo: BoolT  # pyright: ignore[reportGeneralTypeIssues] # Generics are actually erased, but the type is passed through pydantic metadata.
        bar: bool

        baz: BoolT  # pyright: ignore[reportGeneralTypeIssues] # Generics are actually erased, but the type is passed through pydantic metadata.

    assert_models_are_equal(schemas["2000-01-01"][GenericSchema], ExpectedSchema)


def test__schema_field_didnt_exist__field_is_typevar(create_runtime_schemas: CreateRuntimeSchemas):
    schemas = create_runtime_schemas(
        version_change(
            schema(GenericSchema).field("foo").didnt_exist,
        ),
    )

    class ExpectedSchema(BaseModel):
        bar: bool

    assert_models_are_equal(schemas["2000-01-01"][GenericSchema], ExpectedSchema)


@pytest.mark.parametrize(*ATTR_MAP)
def test__schema_field_had__modifying_typevar_field(
    attr: PossibleFieldAttributes,
    value: Any,
    create_runtime_schemas: CreateRuntimeSchemas,
):
    assert_field_had_changes_apply(
        GenericSchema,
        attr,
        value,
        create_runtime_schemas,
    )


@pytest.mark.parametrize(*ATTR_MAP)
def test__schema_field_didnt_have__modifying_typevar_field(
    attr: PossibleFieldAttributes, value: Any, create_runtime_schemas: CreateRuntimeSchemas
):
    class GenericSchemaWithMetadata(BaseModel, Generic[BoolT]):
        foo: Annotated[BoolT, Field(**dict(ATTR_MAP[1]))]
        bar: bool

    schemas = create_runtime_schemas(
        version_change(
            schema(GenericSchemaWithMetadata).field("foo").didnt_have(attr),
        )
    )

    assert not schemas["2000-01-01"][GenericSchemaWithMetadata].model_fields["foo"].metadata


def test__schema_field_had__field_type_is_narrowed(
    create_runtime_schemas: CreateRuntimeSchemas,
):
    schemas = create_runtime_schemas(
        version_change(
            schema(GenericSchema)
            .field("bar")
            .had(
                type=BoolT,
            ),
        ),
    )

    class ExpectedSchema(BaseModel):
        foo: BoolT  # pyright: ignore[reportGeneralTypeIssues] # Generics are actually erased, but the type is passed through pydantic metadata.
        bar: BoolT  # pyright: ignore[reportGeneralTypeIssues] # Generics are actually erased, but the type is passed through pydantic metadata.

    assert_models_are_equal(schemas["2000-01-01"][GenericSchema], ExpectedSchema)


################
# PARAMETRIZED #
################


def test__schema_field_didnt_exist__parametrized_field_is_missing__should_raise_error(
    create_runtime_schemas: CreateRuntimeSchemas,
):
    with pytest.raises(
        InvalidGenerationInstructionError,
        match=re.escape(
            'You tried to delete a field "foo" from "ParametrizedSchema" '
            'in "MyVersionChange" but it doesn\'t have such a field.',
        ),
    ):
        create_runtime_schemas(
            version_change(
                schema(ParametrizedSchema).field("foo").didnt_exist,
            ),
        )


@pytest.mark.parametrize(*ATTR_MAP)
def test__schema_field_had__modifying_parametrized_field(
    attr: str,
    value: Any,
    create_runtime_schemas: CreateRuntimeSchemas,
):
    assert_field_had_changes_apply(
        ParametrizedSchema,
        attr,
        value,
        create_runtime_schemas,
    )


def test__schema_field_had__field_parametrized_type_is_replaced(create_runtime_schemas: CreateRuntimeSchemas):
    schemas = create_runtime_schemas(
        version_change(
            schema(ParametrizedSchema)
            .field("foo")
            .had(
                type=int,
            ),
        ),
    )

    class ExpectedSchema(GenericSchema[int]):
        pass

    assert_models_are_equal(schemas["2000-01-01"][ParametrizedSchema], ExpectedSchema)



from datetime import date

import pytest
from pydantic import Field

from cadwyn.exceptions import CadwynError
from cadwyn.schema_generation import migrate_response_body
from cadwyn.structure.data import ResponseInfo, convert_response_to_previous_version_for
from cadwyn.structure.schemas import schema
from cadwyn.structure.versions import Version, VersionBundle
from tests.conftest import version_change
from tests.test_data_migrations import EmptySchema


def test__version_with_date__should_be_converted_to_string():
    assert Version(date(2022, 11, 16)).value == "2022-11-16"


def test__manual_response_migrations__with_version_as_date():
    @convert_response_to_previous_version_for(EmptySchema)
    def response_converter(response: ResponseInfo):
        response.body["amount"] = 83

    version_bundle = VersionBundle(
        Version(
            "2001-01-01",
            version_change(
                schema(EmptySchema).field("name").existed_as(type=str, info=Field(default="Apples")),
                schema(EmptySchema).field("amount").existed_as(type=int),
                convert=response_converter,
            ),
        ),
        Version("2000-01-01"),
    )

    new_response = migrate_response_body(
        version_bundle, EmptySchema, latest_body={"id": "hewwo"}, version=date(2000, 1, 1)
    )
    assert new_response.model_dump() == {
        "name": "Apples",
        "amount": 83,
    }
    assert new_response.model_dump(exclude_unset=True) == {"amount": 83}

    with pytest.raises(CadwynError):
        new_response = migrate_response_body(
            version_bundle, EmptySchema, latest_body={"id": "hewwo"}, version=date(1999, 1, 1)
        )



from pathlib import Path

from mkdocs.config import Config

DOCS_DIR = Path(__file__).parent
PROJECT_ROOT = DOCS_DIR.parent


def on_pre_build(config: Config) -> None:
    """Before the build starts"""
    add_changelog()


def add_changelog() -> None:
    changelog = (PROJECT_ROOT / "CHANGELOG.md").read_text(encoding="utf-8")
    new_file = DOCS_DIR / "home" / "CHANGELOG.md"

    # avoid writing file unless the content has changed to avoid infinite build loop
    if not new_file.is_file() or new_file.read_text(encoding="utf-8") != changelog:
        new_file.write_text(changelog, encoding="utf-8")



import pytest
from fastapi import APIRouter
from starlette.responses import JSONResponse, Response

from cadwyn import Cadwyn
from cadwyn.structure.versions import Version, VersionBundle

router = APIRouter(prefix="/v1")


@router.get("/")
def homepage():
    return Response("Hello, world", media_type="text/plain")


@router.get("/users/{username}/{page}")
def users_api(username: str, page: int):
    return JSONResponse(
        {"users": [{"username": username, "page": page}]},
    )


@router.get("/users")
def users():
    return Response("All users", media_type="text/plain")


@router.get("/doggies/{dogname}")
def doggies_api(dogname: str):
    return JSONResponse({"doggies": [{"dogname": dogname}]})


versions = [
    "2022-01-10",
    "2022-02-11",
    "1998-11-15",
    "2022-03-12",
    "2027-11-15",
    "2022-04-14",
]
mixed_hosts_app = Cadwyn(versions=VersionBundle(Version("1998-11-15")))
for version in versions:
    with pytest.warns(DeprecationWarning):
        mixed_hosts_app.add_header_versioned_routers(  # pyright: ignore[reportDeprecated]
            router,
            header_value=version,
        )



from datetime import date
from typing import Annotated

import typer
from rich.console import Console
from rich.syntax import Syntax

from cadwyn._render import render_model_by_path, render_module_by_path

_CONSOLE = Console()
_RAW_ARG = Annotated[bool, typer.Option(help="Output code without color")]

app = typer.Typer(
    name="cadwyn",
    add_completion=False,
    help="Modern Stripe-like API versioning in FastAPI",
)

render_subapp = typer.Typer(
    name="render",
    add_completion=False,
    help="Render pydantic models and enums from a certainn version and output them to stdout",
)

app.add_typer(render_subapp)


def version_callback(value: bool):
    if value:
        from . import __version__

        typer.echo(f"Cadwyn {__version__}")
        raise typer.Exit


def output_code(code: str, raw: bool):
    if raw:
        typer.echo(code)
    else:  # pragma: no cover
        _CONSOLE.print(Syntax(code, "python", line_numbers=True))


@render_subapp.command(
    name="model",
    help="Render a concrete pydantic model or enum from a certain version and output it to stdout",
    short_help="Render a single model or enum",
)
def render(
    model: Annotated[str, typer.Argument(metavar="<module>:<attribute>", help="Python path to the model to render")],
    app: Annotated[str, typer.Option(metavar="<module>:<attribute>", help="Python path to the main Cadwyn app")],
    version: Annotated[str, typer.Option(parser=lambda s: str(date.fromisoformat(s)), metavar="ISO-VERSION")],
    raw: _RAW_ARG = False,
) -> None:
    output_code(render_model_by_path(model, app, version), raw)


@render_subapp.command(
    name="module",
    help="Render all versioned models and enums within a module from a certain version and output them to stdout",
    short_help="Render all models and enums from an entire module",
)
def render_module(
    module: Annotated[str, typer.Argument(metavar="<module>", help="Python path to the module to render")],
    app: Annotated[str, typer.Option(metavar="<module>:<attribute>", help="Python path to the main Cadwyn app")],
    version: Annotated[str, typer.Option(parser=lambda s: str(date.fromisoformat(s)), metavar="ISO-VERSION")],
    raw: _RAW_ARG = False,
) -> None:
    output_code(render_module_by_path(module, app, version), raw)


@app.callback()
def main(
    version: bool = typer.Option(None, "-V", "--version", callback=version_callback, is_eager=True),
): ...


if __name__ == "__main__":
    app()



from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, Union, cast

from pydantic import AliasChoices, AliasPath, BaseModel, Field
from pydantic._internal._decorators import PydanticDescriptorProxy, unwrap_wrapped_function
from pydantic.fields import FieldInfo

from cadwyn._utils import (
    DATACLASS_SLOTS,
    Sentinel,
    fully_unwrap_decorator,
    get_name_of_function_wrapped_in_pydantic_validator,
    lenient_issubclass,
)
from cadwyn.exceptions import CadwynStructureError

from .common import _HiddenAttributeMixin

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny


PossibleFieldAttributes = Literal[
    "default",
    "alias",
    "alias_priority",
    "default_factory",
    "validation_alias",
    "serialization_alias",
    "title",
    "field_title_generator",
    "description",
    "examples",
    "exclude",
    "const",
    "deprecated",
    "frozen",
    "validate_default",
    "repr",
    "init",
    "init_var",
    "kw_only",
    "fail_fast",
    "gt",
    "ge",
    "lt",
    "le",
    "strict",
    "coerce_numbers_to_str",
    "multiple_of",
    "allow_inf_nan",
    "max_digits",
    "decimal_places",
    "min_length",
    "max_length",
    "union_mode",
    "allow_mutation",
    "pattern",
    "discriminator",
    "json_schema_extra",
]


# TODO: Add json_schema_extra as a breaking change in a major version
@dataclass(**DATACLASS_SLOTS)
class FieldChanges:
    default: Any
    alias: Union[str, None]
    default_factory: Any
    alias_priority: Union[int, None]
    validation_alias: Union[str, AliasPath, AliasChoices, None]
    serialization_alias: Union[str, None]
    title: Union[str, None]
    field_title_generator: Union[Callable[[str, FieldInfo], str], None]
    description: str
    examples: Union[list[Any], None]
    exclude: "AbstractSetIntStr | MappingIntStrAny | Any"
    const: bool
    deprecated: bool
    frozen: Union[bool, None]
    validate_default: Union[bool, None]
    repr: bool
    init: Union[bool, None]
    init_var: Union[bool, None]
    kw_only: Union[bool, None]
    fail_fast: bool
    gt: float
    ge: float
    lt: float
    le: float
    strict: bool
    coerce_numbers_to_str: Union[bool, None]
    multiple_of: float
    allow_inf_nan: bool
    max_digits: int
    decimal_places: int
    min_length: int
    max_length: int
    union_mode: Literal["smart", "left_to_right"]
    allow_mutation: bool
    pattern: str
    discriminator: str
    json_schema_extra: Union[dict[str, Any], Callable[[dict[str, Any]], None], None]


@dataclass(**DATACLASS_SLOTS)
class FieldHadInstruction(_HiddenAttributeMixin):
    schema: type[BaseModel]
    name: str
    type: type
    field_changes: FieldChanges
    new_name: str


@dataclass(**DATACLASS_SLOTS)
class FieldDidntHaveInstruction(_HiddenAttributeMixin):
    schema: type[BaseModel]
    name: str
    attributes: tuple[str, ...]


@dataclass(**DATACLASS_SLOTS)
class FieldDidntExistInstruction(_HiddenAttributeMixin):
    schema: type[BaseModel]
    name: str


@dataclass(**DATACLASS_SLOTS)
class FieldExistedAsInstruction(_HiddenAttributeMixin):
    schema: type[BaseModel]
    name: str
    field: FieldInfo


# TODO (https://github.com/zmievsa/cadwyn/issues/112): Add an ability to add extras
@dataclass(**DATACLASS_SLOTS)
class AlterFieldInstructionFactory:
    schema: type[BaseModel]
    name: str

    def had(
        self,
        *,
        name: str = Sentinel,
        type: Any = Sentinel,
        default: Any = Sentinel,
        alias: Union[str, None] = Sentinel,
        default_factory: Callable = Sentinel,
        alias_priority: Union[int, None] = Sentinel,
        validation_alias: Union[str, AliasPath, AliasChoices, None] = Sentinel,
        serialization_alias: Union[str, None] = Sentinel,
        title: Union[str, None] = Sentinel,
        field_title_generator: Union[Callable[[str, FieldInfo], str], None] = Sentinel,
        description: str = Sentinel,
        examples: Union[list[Any], None] = Sentinel,
        exclude: "AbstractSetIntStr | MappingIntStrAny | Any" = Sentinel,
        const: bool = Sentinel,
        deprecated: bool = Sentinel,
        frozen: Union[bool, None] = Sentinel,
        validate_default: Union[bool, None] = Sentinel,
        repr: bool = Sentinel,
        init: bool = Sentinel,
        init_var: bool = Sentinel,
        kw_only: bool = Sentinel,
        fail_fast: bool = Sentinel,
        gt: float = Sentinel,
        ge: float = Sentinel,
        lt: float = Sentinel,
        le: float = Sentinel,
        strict: bool = Sentinel,
        coerce_numbers_to_str: bool = Sentinel,
        multiple_of: float = Sentinel,
        allow_inf_nan: bool = Sentinel,
        max_digits: int = Sentinel,
        decimal_places: int = Sentinel,
        min_length: int = Sentinel,
        max_length: int = Sentinel,
        union_mode: Literal["smart", "left_to_right"] = Sentinel,
        allow_mutation: bool = Sentinel,
        pattern: str = Sentinel,
        discriminator: str = Sentinel,
        json_schema_extra: Union[dict[str, Any], Callable[[dict[str, Any]], None], None] = Sentinel,
    ) -> FieldHadInstruction:
        return FieldHadInstruction(
            is_hidden_from_changelog=False,
            schema=self.schema,
            name=self.name,
            type=type,
            new_name=name,
            field_changes=FieldChanges(
                default=default,
                default_factory=default_factory,
                alias_priority=alias_priority,
                alias=alias,
                validation_alias=validation_alias,
                serialization_alias=serialization_alias,
                title=title,
                field_title_generator=field_title_generator,
                description=description,
                examples=examples,
                exclude=exclude,
                const=const,
                deprecated=deprecated,
                frozen=frozen,
                validate_default=validate_default,
                repr=repr,
                init=init,
                init_var=init_var,
                kw_only=kw_only,
                fail_fast=fail_fast,
                gt=gt,
                ge=ge,
                lt=lt,
                le=le,
                strict=strict,
                coerce_numbers_to_str=coerce_numbers_to_str,
                multiple_of=multiple_of,
                allow_inf_nan=allow_inf_nan,
                max_digits=max_digits,
                decimal_places=decimal_places,
                min_length=min_length,
                max_length=max_length,
                union_mode=union_mode,
                allow_mutation=allow_mutation,
                pattern=pattern,
                discriminator=discriminator,
                json_schema_extra=json_schema_extra,
            ),
        )

    def didnt_have(self, *attributes: PossibleFieldAttributes) -> FieldDidntHaveInstruction:
        for attribute in attributes:
            if attribute not in FieldChanges.__dataclass_fields__:
                raise CadwynStructureError(
                    f"Unknown attribute {attribute!r}. Are you sure it's a valid field attribute?"
                )
        return FieldDidntHaveInstruction(
            is_hidden_from_changelog=False, schema=self.schema, name=self.name, attributes=attributes
        )

    @property
    def didnt_exist(self) -> FieldDidntExistInstruction:
        return FieldDidntExistInstruction(is_hidden_from_changelog=False, schema=self.schema, name=self.name)

    def existed_as(
        self,
        *,
        type: Any,
        info: Union[FieldInfo, Any, None] = None,
    ) -> FieldExistedAsInstruction:
        if info is None:
            info = cast("FieldInfo", Field())
        info.annotation = type
        return FieldExistedAsInstruction(is_hidden_from_changelog=False, schema=self.schema, name=self.name, field=info)


def _get_model_decorators(model: type[BaseModel]):
    return [
        *model.__pydantic_decorators__.validators.values(),
        *model.__pydantic_decorators__.field_validators.values(),
        *model.__pydantic_decorators__.root_validators.values(),
        *model.__pydantic_decorators__.field_serializers.values(),
        *model.__pydantic_decorators__.model_serializers.values(),
        *model.__pydantic_decorators__.model_validators.values(),
        *model.__pydantic_decorators__.computed_fields.values(),
    ]


@dataclass(**DATACLASS_SLOTS)
class ValidatorExistedInstruction:
    schema: type[BaseModel]
    validator: Union[Callable[..., Any], PydanticDescriptorProxy]


@dataclass(**DATACLASS_SLOTS)
class ValidatorDidntExistInstruction:
    schema: type[BaseModel]
    name: str


@dataclass(**DATACLASS_SLOTS)
class AlterValidatorInstructionFactory:
    schema: type[BaseModel]
    func: Union[Callable[..., Any], PydanticDescriptorProxy]

    @property
    def existed(self) -> ValidatorExistedInstruction:
        return ValidatorExistedInstruction(self.schema, self.func)

    @property
    def didnt_exist(self) -> ValidatorDidntExistInstruction:
        return ValidatorDidntExistInstruction(
            self.schema, get_name_of_function_wrapped_in_pydantic_validator(self.func)
        )


AlterSchemaSubInstruction = Union[
    FieldHadInstruction,
    FieldDidntHaveInstruction,
    FieldDidntExistInstruction,
    FieldExistedAsInstruction,
    ValidatorExistedInstruction,
    ValidatorDidntExistInstruction,
]


@dataclass(**DATACLASS_SLOTS)
class SchemaHadInstruction(_HiddenAttributeMixin):
    schema: type[BaseModel]
    name: str


@dataclass(**DATACLASS_SLOTS)
class AlterSchemaInstructionFactory:
    schema: type[BaseModel]

    def field(self, name: str, /) -> AlterFieldInstructionFactory:
        return AlterFieldInstructionFactory(self.schema, name)

    def validator(
        self, func: "Union[Callable[..., Any], classmethod[Any, Any, Any], PydanticDescriptorProxy]", /
    ) -> AlterValidatorInstructionFactory:
        func = cast("Union[Callable[..., Any], PydanticDescriptorProxy]", unwrap_wrapped_function(func))

        if not isinstance(func, PydanticDescriptorProxy):
            if hasattr(func, "__self__"):
                owner = func.__self__
                if lenient_issubclass(owner, BaseModel) and any(  # pragma: no branch
                    fully_unwrap_decorator(decorator.func, decorator.shim) == func
                    for decorator in _get_model_decorators(owner)
                ):
                    return AlterValidatorInstructionFactory(self.schema, func)
            raise CadwynStructureError("The passed function must be a pydantic validator")
        return AlterValidatorInstructionFactory(self.schema, func)

    def had(self, *, name: str) -> SchemaHadInstruction:
        return SchemaHadInstruction(is_hidden_from_changelog=False, schema=self.schema, name=name)


def schema(model: type[BaseModel], /) -> AlterSchemaInstructionFactory:
    return AlterSchemaInstructionFactory(model)



import functools
import inspect
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import ClassVar, Union, cast

from fastapi import Request, Response
from starlette.datastructures import FormData, MutableHeaders, UploadFile
from typing_extensions import Any, ParamSpec, overload

from cadwyn._utils import same_definition_as_in
from cadwyn.structure.endpoints import _validate_that_strings_are_valid_http_methods

_P = ParamSpec("_P")


class RequestInfo:
    __slots__ = ("_cookies", "_form", "_query_params", "_request", "body", "headers")

    def __init__(self, request: Request, body: Any):
        super().__init__()
        self.body = body
        self.headers = request.headers.mutablecopy()
        self._cookies = request.cookies
        self._query_params = request.query_params._dict
        self._request = request
        if isinstance(body, FormData):
            self._form: Union[list[tuple[str, Union[UploadFile, str]]], None] = list(body.multi_items())
        else:
            self._form = None

    @property
    def cookies(self) -> dict[str, str]:
        return self._cookies

    @property
    def query_params(self) -> dict[str, str]:
        return self._query_params

    @property
    def form(self) -> Union[list[tuple[str, Union[UploadFile, str]]], None]:
        return self._form


# TODO (https://github.com/zmievsa/cadwyn/issues/111): handle _response.media_type and _response.background
class ResponseInfo:
    __slots__ = ("_response", "body")

    def __init__(self, response: Response, body: Any):
        super().__init__()
        self.body = body
        self._response = response

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @status_code.setter
    def status_code(self, value: int):
        self._response.status_code = value

    @property
    def headers(self) -> MutableHeaders:
        return self._response.headers

    @same_definition_as_in(Response.set_cookie)
    def set_cookie(self, *args: Any, **kwargs: Any):
        return self._response.set_cookie(*args, **kwargs)

    @same_definition_as_in(Response.delete_cookie)
    def delete_cookie(self, *args: Any, **kwargs: Any):
        return self._response.delete_cookie(*args, **kwargs)


@dataclass
class _AlterDataInstruction:
    transformer: Callable[[Any], None]
    owner: type = field(init=False)
    _payload_arg_name: ClassVar[str]

    def __post_init__(self):
        signature = inspect.signature(self.transformer)
        if list(signature.parameters) != [self._payload_arg_name]:
            raise ValueError(
                f"Method '{self.transformer.__name__}' must have only 1 parameter: {self._payload_arg_name}",
            )

        functools.update_wrapper(self, self.transformer)

    def __set_name__(self, owner: type, name: str):
        self.owner = owner

    def __call__(self, __request_or_response: Union[RequestInfo, ResponseInfo], /) -> None:
        return self.transformer(__request_or_response)


@dataclass
class _BaseAlterBySchemaInstruction:
    schemas: tuple[Any, ...]
    check_usage: bool = True


##########
# Requests
##########


@dataclass
class _BaseAlterRequestInstruction(_AlterDataInstruction):
    _payload_arg_name = "request"


@dataclass
class _AlterRequestBySchemaInstruction(_BaseAlterBySchemaInstruction, _BaseAlterRequestInstruction): ...


@dataclass
class _AlterRequestByPathInstruction(_BaseAlterRequestInstruction):
    path: str
    methods: set[str]
    repr_name = "Request by path converter"


@overload
def convert_request_to_next_version_for(
    first_schema: type,
    /,
    *additional_schemas: type,
    check_usage: bool = True,
) -> "type[staticmethod[_P, None]]": ...


@overload
def convert_request_to_next_version_for(path: str, methods: list[str], /) -> "type[staticmethod[_P, None]]": ...


def convert_request_to_next_version_for(
    schema_or_path: Union[type, str],
    methods_or_second_schema: Union[list[str], None, type] = None,
    /,
    *additional_schemas: type,
    check_usage: bool = True,
) -> "type[staticmethod[_P, None]]":
    _validate_decorator_args(schema_or_path, methods_or_second_schema, additional_schemas)

    def decorator(transformer: Callable[[RequestInfo], None]) -> Any:
        if isinstance(schema_or_path, str):
            return _AlterRequestByPathInstruction(
                path=schema_or_path,
                methods=set(cast("list", methods_or_second_schema)),
                transformer=transformer,
            )
        else:
            if methods_or_second_schema is None:
                schemas = (schema_or_path,)
            else:
                schemas = (schema_or_path, methods_or_second_schema, *additional_schemas)
            return _AlterRequestBySchemaInstruction(
                schemas=schemas,
                transformer=transformer,
                check_usage=check_usage,
            )

    return decorator  # pyright: ignore[reportReturnType]


###########
# Responses
###########


@dataclass
class _BaseAlterResponseInstruction(_AlterDataInstruction):
    _payload_arg_name = "response"
    migrate_http_errors: bool


@dataclass
class _AlterResponseBySchemaInstruction(_BaseAlterBySchemaInstruction, _BaseAlterResponseInstruction): ...


@dataclass
class _AlterResponseByPathInstruction(_BaseAlterResponseInstruction):
    path: str
    methods: set[str]
    repr_name = "Response by path converter"


@overload
def convert_response_to_previous_version_for(
    first_schema: type,
    /,
    *schemas: type,
    migrate_http_errors: bool = False,
    check_usage: bool = True,
) -> "type[staticmethod[_P, None]]": ...


@overload
def convert_response_to_previous_version_for(
    path: str,
    methods: list[str],
    /,
    *,
    migrate_http_errors: bool = False,
) -> "type[staticmethod[_P, None]]": ...


def convert_response_to_previous_version_for(
    schema_or_path: Union[type, str],
    methods_or_second_schema: Union[list[str], type, None] = None,
    /,
    *additional_schemas: type,
    migrate_http_errors: bool = False,
    check_usage: bool = True,
) -> "type[staticmethod[_P, None]]":
    _validate_decorator_args(schema_or_path, methods_or_second_schema, additional_schemas)

    def decorator(transformer: Callable[[ResponseInfo], None]) -> Any:
        if isinstance(schema_or_path, str):
            # The validation above checks that methods is not None
            return _AlterResponseByPathInstruction(
                path=schema_or_path,
                methods=set(cast("list", methods_or_second_schema)),
                transformer=transformer,
                migrate_http_errors=migrate_http_errors,
            )
        else:
            if methods_or_second_schema is None:
                schemas = (schema_or_path,)
            else:
                schemas = (schema_or_path, methods_or_second_schema, *additional_schemas)
            return _AlterResponseBySchemaInstruction(
                schemas=schemas,
                transformer=transformer,
                migrate_http_errors=migrate_http_errors,
                check_usage=check_usage,
            )

    return decorator  # pyright: ignore[reportReturnType]


def _validate_decorator_args(
    schema_or_path: Union[type, str],
    methods_or_second_schema: Union[list[str], type, None],
    additional_schemas: tuple[type, ...],
) -> None:
    if isinstance(schema_or_path, str):
        if not isinstance(methods_or_second_schema, list):
            raise TypeError("If path was provided as a first argument, methods must be provided as a second argument")
        _validate_that_strings_are_valid_http_methods(methods_or_second_schema)
        if additional_schemas:
            raise TypeError("If path was provided as a first argument, then additional schemas cannot be added")

    elif methods_or_second_schema is not None and not isinstance(methods_or_second_schema, type):
        raise TypeError("If schema was provided as a first argument, all other arguments must also be schemas")



import re
from collections import defaultdict
from collections.abc import Callable, Sequence
from copy import copy, deepcopy
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Union,
    cast,
)

import fastapi.params
import fastapi.routing
import fastapi.security.base
import fastapi.utils
from fastapi import APIRouter
from fastapi.routing import APIRoute
from pydantic import BaseModel
from starlette.routing import BaseRoute
from typing_extensions import TypeVar, assert_never

from cadwyn._utils import DATACLASS_SLOTS, Sentinel, lenient_issubclass
from cadwyn.exceptions import (
    CadwynError,
    RouteAlreadyExistsError,
    RouteByPathConverterDoesNotApplyToAnythingError,
    RouteRequestBySchemaConverterDoesNotApplyToAnythingError,
    RouteResponseBySchemaConverterDoesNotApplyToAnythingError,
    RouterGenerationError,
    RouterPathParamsModifiedError,
)
from cadwyn.schema_generation import (
    _add_request_and_response_params,
    generate_versioned_models,
)
from cadwyn.structure import Version, VersionBundle
from cadwyn.structure.common import Endpoint, VersionType
from cadwyn.structure.data import _AlterRequestByPathInstruction, _AlterResponseByPathInstruction
from cadwyn.structure.endpoints import (
    EndpointDidntExistInstruction,
    EndpointExistedInstruction,
    EndpointHadInstruction,
)
from cadwyn.structure.versions import VersionChange

if TYPE_CHECKING:
    from fastapi.dependencies.models import Dependant

    from cadwyn.middleware import APIVersionLocation

_Call = TypeVar("_Call", bound=Callable[..., Any])
_R = TypeVar("_R", bound=APIRouter)
_WR = TypeVar("_WR", bound=APIRouter, default=APIRouter)
_RouteT = TypeVar("_RouteT", bound=BaseRoute)
# This is a hack we do because we can't guarantee how the user will use the router.
_DELETED_ROUTE_TAG = "_CADWYN_DELETED_ROUTE"
_RoutePath = str
_RouteMethod = str
_RouteId = int


@dataclass(**DATACLASS_SLOTS, frozen=True, eq=True)
class _EndpointInfo:
    endpoint_path: str
    endpoint_methods: frozenset[str]


@dataclass(**DATACLASS_SLOTS, frozen=True)
class GeneratedRouters(Generic[_R, _WR]):
    endpoints: dict[VersionType, _R]
    webhooks: dict[VersionType, _WR]


def generate_versioned_routers(
    router: _R,
    versions: VersionBundle,
    *,
    webhooks: Union[_WR, None] = None,
    api_version_parameter_name: str = "X-API-Version",
    api_version_location: "APIVersionLocation" = "custom_header",
) -> GeneratedRouters[_R, _WR]:
    if webhooks is None:
        webhooks = cast("_WR", APIRouter())
    return _EndpointTransformer(
        router, versions, webhooks, api_version_parameter_name, api_version_location
    ).transform()


class VersionedAPIRouter(fastapi.routing.APIRouter):
    def only_exists_in_older_versions(self, endpoint: _Call) -> _Call:
        route = _get_route_from_func(self.routes, endpoint)
        if route is None:
            raise LookupError(
                f'Route not found on endpoint: "{endpoint.__name__}". '
                "Are you sure it's a route and decorators are in the correct order?",
            )
        if _DELETED_ROUTE_TAG in route.tags:
            raise CadwynError(f'The route "{endpoint.__name__}" was already deleted. You can\'t delete it again.')
        route.tags.append(_DELETED_ROUTE_TAG)
        return endpoint


def copy_router(router: _R) -> _R:
    router = copy(router)
    router.routes = [copy_route(r) for r in router.routes]
    return router


def copy_route(route: _RouteT) -> _RouteT:
    if not isinstance(route, APIRoute):
        return copy(route)

    # This is slightly wasteful in terms of resources but it makes it easy for us
    # to make sure that new versions of FastAPI are going to be supported even if
    # APIRoute gets new attributes.
    # Pre-populate deepcopy's memo so it returns the original objects for attributes
    # containing ModelField/TypeAdapter instances instead of recursing into them.
    # These can hold TypeAdapters for recursive types (e.g. JsonValue) that cause
    # infinite recursion during deepcopy.
    memo: dict[int, Any] = {}
    for attr in ("dependant", "_flat_dependant", "body_field"):
        obj = getattr(route, attr, None)
        if obj is not None:
            memo[id(obj)] = obj
    new_route = deepcopy(route, memo)
    new_route.dependant = copy(route.dependant)
    if getattr(route, "_flat_dependant", None) is not None:
        new_route._flat_dependant = copy(route._flat_dependant)
    new_route.body_field = route.body_field
    new_route.dependencies = copy(route.dependencies)
    return new_route


class _EndpointTransformer(Generic[_R, _WR]):
    def __init__(
        self,
        parent_router: _R,
        versions: VersionBundle,
        webhooks: _WR,
        api_version_parameter_name: str,
        api_version_location: "APIVersionLocation",
    ) -> None:
        super().__init__()
        self.parent_router = parent_router
        self.versions = versions
        self.parent_webhooks_router = webhooks
        self.api_version_parameter_name = api_version_parameter_name
        self.api_version_location: APIVersionLocation = api_version_location
        self.schema_generators = generate_versioned_models(versions)

        self.routes_that_never_existed = [
            route for route in parent_router.routes if isinstance(route, APIRoute) and _DELETED_ROUTE_TAG in route.tags
        ]

    def transform(self) -> GeneratedRouters[_R, _WR]:
        # Copy MUST keep the order and number of routes. Otherwise, a ton of code below will break.
        router = copy_router(self.parent_router)
        webhook_router = copy_router(self.parent_webhooks_router)
        routers: dict[VersionType, _R] = {}
        webhook_routers: dict[VersionType, _WR] = {}

        for version in self.versions:
            self.schema_generators[str(version.value)].annotation_transformer.migrate_router_to_version(router)
            self.schema_generators[str(version.value)].annotation_transformer.migrate_router_to_version(webhook_router)

            self._attach_routes_to_data_converters(router, self.parent_router, version)

            routers[version.value] = router
            webhook_routers[version.value] = webhook_router
            # Applying changes for the next version
            router = copy_router(router)
            webhook_router = copy_router(webhook_router)
            self._apply_endpoint_changes_to_router(router.routes + webhook_router.routes, version)

        if self.routes_that_never_existed:
            raise RouterGenerationError(
                "Every route you mark with "
                f"@VersionedAPIRouter.{VersionedAPIRouter.only_exists_in_older_versions.__name__} "
                "must be restored in one of the older versions. Otherwise you just need to delete it altogether. "
                "The following routes have been marked with that decorator but were never restored: "
                f"{self.routes_that_never_existed}",
            )

        for route_index, head_route in enumerate(self.parent_router.routes):
            if not isinstance(head_route, APIRoute):
                continue
            _add_request_and_response_params(head_route)
            copy_of_dependant = copy(head_route.dependant)

            for older_router in list(routers.values()):
                older_route = older_router.routes[route_index]

                # We know they are APIRoutes because of the check at the very beginning of the top loop.
                # I.e. Because head_route is an APIRoute, both routes are  APIRoutes too
                older_route = cast("APIRoute", older_route)
                # Wait.. Why do we need this code again?
                if older_route.body_field is not None and _route_has_a_simple_body_schema(older_route):
                    annotation = older_route.body_field.field_info.annotation
                    if hasattr(annotation, "__cadwyn_original_model__"):
                        template_older_body_model = annotation.__cadwyn_original_model__  # pyright: ignore[reportOptionalMemberAccess]
                    else:
                        template_older_body_model = annotation
                else:
                    template_older_body_model = None
                _add_data_migrations_to_route(
                    older_route,
                    # NOTE: The fact that we use latest here assumes that the route can never change its response schema
                    head_route,
                    template_older_body_model,
                    older_route.body_field.alias if older_route.body_field is not None else None,
                    copy_of_dependant,
                    self.versions,
                )
        for router in routers.values():
            router.routes = [
                route
                for route in router.routes
                if not (isinstance(route, fastapi.routing.APIRoute) and _DELETED_ROUTE_TAG in route.tags)
            ]
        for webhook_router in webhook_routers.values():
            webhook_router.routes = [
                route
                for route in webhook_router.routes
                if not (isinstance(route, fastapi.routing.APIRoute) and _DELETED_ROUTE_TAG in route.tags)
            ]
        return GeneratedRouters(routers, webhook_routers)

    def _attach_routes_to_data_converters(self, router: APIRouter, head_router: APIRouter, version: Version):
        # This method is way out of its league in terms of complexity. We gotta refactor it.

        path_to_route_methods_mapping, head_response_models, head_request_bodies = (
            self._extract_all_routes_identifiers_for_route_to_converter_matching(router)
        )

        for version_change in version.changes:
            for by_path_converters in [
                *version_change.alter_response_by_path_instructions.values(),
                *version_change.alter_request_by_path_instructions.values(),
            ]:
                for by_path_converter in by_path_converters:
                    self._attach_routes_by_path_converter(
                        head_router, path_to_route_methods_mapping, version_change, by_path_converter
                    )

            for by_schema_converters in version_change.alter_request_by_schema_instructions.values():
                for by_schema_converter in by_schema_converters:
                    if not by_schema_converter.check_usage:  # pragma: no cover
                        continue
                    missing_models = set(by_schema_converter.schemas) - head_request_bodies
                    if missing_models:
                        raise RouteRequestBySchemaConverterDoesNotApplyToAnythingError(
                            f"Request by body schema converter "
                            f'"{version_change.__name__}.{by_schema_converter.transformer.__name__}" '
                            f"failed to find routes with the following body schemas: "
                            f"{[m.__name__ for m in missing_models]}. "
                            f"This means that you are trying to apply this converter to non-existing endpoint(s). "
                        )
            for by_schema_converters in version_change.alter_response_by_schema_instructions.values():
                for by_schema_converter in by_schema_converters:
                    if not by_schema_converter.check_usage:  # pragma: no cover
                        continue
                    missing_models = set(by_schema_converter.schemas) - head_response_models
                    if missing_models:
                        raise RouteResponseBySchemaConverterDoesNotApplyToAnythingError(
                            f"Response by response model converter "
                            f'"{version_change.__name__}.{by_schema_converter.transformer.__name__}" '
                            f"failed to find routes with the following response models: "
                            f"{[m.__name__ for m in missing_models]}. "
                            f"This means that you are trying to apply this converter to non-existing endpoint(s). "
                            "If this is intentional and this converter really does not apply to any endpoints, then "
                            "pass check_usage=False argument to "
                            f"{version_change.__name__}.{by_schema_converter.transformer.__name__}"
                        )

    def _attach_routes_by_path_converter(
        self,
        head_router: APIRouter,
        path_to_route_methods_mapping: dict[_RoutePath, dict[_RouteMethod, set[_RouteId]]],
        version_change: type[VersionChange],
        by_path_converter: Union[_AlterResponseByPathInstruction, _AlterRequestByPathInstruction],
    ):
        missing_methods = set()
        for method in by_path_converter.methods:
            if method in path_to_route_methods_mapping[by_path_converter.path]:
                for route_index in path_to_route_methods_mapping[by_path_converter.path][method]:
                    route = head_router.routes[route_index]
                    if isinstance(by_path_converter, _AlterResponseByPathInstruction):
                        version_change._route_to_response_migration_mapping[id(route)].append(by_path_converter)
                    else:
                        version_change._route_to_request_migration_mapping[id(route)].append(by_path_converter)
            else:
                missing_methods.add(method)

        if missing_methods:
            raise RouteByPathConverterDoesNotApplyToAnythingError(
                f"{by_path_converter.repr_name} "
                f'"{version_change.__name__}.{by_path_converter.transformer.__name__}" '
                f"failed to find routes with the following methods: {list(missing_methods)}. "
                f"This means that you are trying to apply this converter to non-existing endpoint(s). "
                "Please, check whether the path and methods are correct. (hint: path must include "
                "all path variables and have a name that was used in the version that this "
                "VersionChange resides in)"
            )

    def _extract_all_routes_identifiers_for_route_to_converter_matching(
        self, router: APIRouter
    ) -> tuple[dict[_RoutePath, dict[_RouteMethod, set[_RouteId]]], set[Any], set[Any]]:
        # int is the index of the route in the router.routes list.
        # So we essentially keep track of which routes have which response models and request bodies.
        # and their indices in the router.routes list. The indices will allow us to match them to the same
        # routes in the head version. This gives us the ability to later apply changes to these routes
        # without thinking about any renamings or response model changes.

        response_models = set()
        request_bodies = set()
        path_to_route_methods_mapping: dict[str, dict[str, set[int]]] = defaultdict(lambda: defaultdict(set))

        for index, route in enumerate(router.routes):
            if isinstance(route, APIRoute):
                if route.response_model is not None and lenient_issubclass(route.response_model, BaseModel):
                    response_models.add(route.response_model)
                    # Not sure if it can ever be None when it's a simple schema. Eh, I would rather be safe than sorry
                if _route_has_a_simple_body_schema(route) and route.body_field is not None:
                    annotation = route.body_field.field_info.annotation
                    if annotation is not None and lenient_issubclass(annotation, BaseModel):
                        request_bodies.add(annotation)
                for method in route.methods:
                    path_to_route_methods_mapping[route.path][method].add(index)

        head_response_models = {model.__cadwyn_original_model__ for model in response_models}
        head_request_bodies = {getattr(body, "__cadwyn_original_model__", body) for body in request_bodies}

        return path_to_route_methods_mapping, head_response_models, head_request_bodies

    # TODO (https://github.com/zmievsa/cadwyn/issues/28): Simplify
    def _apply_endpoint_changes_to_router(  # noqa: C901
        self,
        routes: Union[list[BaseRoute], list[APIRoute]],
        version: Version,
    ):
        for version_change in version.changes:
            for instruction in version_change.alter_endpoint_instructions:
                original_routes = _get_routes(
                    routes,
                    instruction.endpoint_path,
                    instruction.endpoint_methods,
                    instruction.endpoint_func_name,
                    is_deleted=False,
                )
                methods_to_which_we_applied_changes = set()
                methods_we_should_have_applied_changes_to = instruction.endpoint_methods.copy()

                if isinstance(instruction, EndpointDidntExistInstruction):
                    deleted_routes = _get_routes(
                        routes,
                        instruction.endpoint_path,
                        instruction.endpoint_methods,
                        instruction.endpoint_func_name,
                        is_deleted=True,
                    )
                    if deleted_routes:
                        method_union = set()
                        for deleted_route in deleted_routes:
                            method_union |= deleted_route.methods
                        raise RouterGenerationError(
                            f'Endpoint "{list(method_union)} {instruction.endpoint_path}" you tried to delete in '
                            f'"{version_change.__name__}" was already deleted in a newer version. If you really have '
                            f'two routes with the same paths and methods, please, use "endpoint(..., func_name=...)" '
                            f"to distinguish between them. Function names of endpoints that were already deleted: "
                            f"{[r.endpoint.__name__ for r in deleted_routes]}",
                        )
                    for original_route in original_routes:
                        methods_to_which_we_applied_changes |= original_route.methods
                        original_route.tags.append(_DELETED_ROUTE_TAG)
                    err = (
                        'Endpoint "{endpoint_methods} {endpoint_path}" you tried to delete in'
                        ' "{version_change_name}" doesn\'t exist in a newer version'
                    )
                elif isinstance(instruction, EndpointExistedInstruction):
                    if original_routes:
                        method_union = set()
                        for original_route in original_routes:
                            method_union |= original_route.methods
                        raise RouterGenerationError(
                            f'Endpoint "{list(method_union)} {instruction.endpoint_path}" you tried to restore in'
                            f' "{version_change.__name__}" already existed in a newer version. If you really have two '
                            f'routes with the same paths and methods, please, use "endpoint(..., func_name=...)" to '
                            f"distinguish between them. Function names of endpoints that already existed: "
                            f"{[r.endpoint.__name__ for r in original_routes]}",
                        )
                    deleted_routes = _get_routes(
                        routes,
                        instruction.endpoint_path,
                        instruction.endpoint_methods,
                        instruction.endpoint_func_name,
                        is_deleted=True,
                    )
                    try:
                        _validate_no_repetitions_in_routes(deleted_routes)
                    except RouteAlreadyExistsError as e:
                        raise RouterGenerationError(
                            f'Endpoint "{list(instruction.endpoint_methods)} {instruction.endpoint_path}" you tried to '
                            f'restore in "{version_change.__name__}" has {len(e.routes)} applicable routes that could '
                            f"be restored. If you really have two routes with the same paths and methods, please, use "
                            f'"endpoint(..., func_name=...)" to distinguish between them. Function names of '
                            f"endpoints that can be restored: {[r.endpoint.__name__ for r in e.routes]}",
                        ) from e
                    for deleted_route in deleted_routes:
                        methods_to_which_we_applied_changes |= deleted_route.methods
                        deleted_route.tags.remove(_DELETED_ROUTE_TAG)

                        routes_that_never_existed = _get_routes(
                            self.routes_that_never_existed,
                            deleted_route.path,
                            deleted_route.methods,
                            deleted_route.endpoint.__name__,
                            is_deleted=True,
                        )
                        if len(routes_that_never_existed) == 1:
                            self.routes_that_never_existed.remove(routes_that_never_existed[0])
                        elif len(routes_that_never_existed) > 1:  # pragma: no cover
                            # I am not sure if it's possible to get to this error but I also don't want
                            # to remove it because I like its clarity very much
                            routes = routes_that_never_existed
                            raise RouterGenerationError(
                                f'Endpoint "{list(deleted_route.methods)} {deleted_route.path}" you tried to restore '
                                f'in "{version_change.__name__}" has {len(routes_that_never_existed)} applicable '
                                f"routes with the same function name and path that could be restored. This can cause "
                                f"problems during version generation. Specifically, Cadwyn won't be able to warn "
                                f"you when you deleted a route and never restored it. Please, make sure that "
                                f"functions for all these routes have different names: "
                                f"{[f'{r.endpoint.__module__}.{r.endpoint.__name__}' for r in routes]}",
                            )
                    err = (
                        'Endpoint "{endpoint_methods} {endpoint_path}" you tried to restore in'
                        ' "{version_change_name}" wasn\'t among the deleted routes'
                    )
                elif isinstance(instruction, EndpointHadInstruction):
                    for original_route in original_routes:
                        methods_to_which_we_applied_changes |= original_route.methods
                        _apply_endpoint_had_instruction(
                            version_change.__name__,
                            instruction,
                            original_route,
                            api_version_parameter_name=self.api_version_parameter_name,
                            api_version_location=self.api_version_location,
                        )
                    err = (
                        'Endpoint "{endpoint_methods} {endpoint_path}" you tried to change in'
                        ' "{version_change_name}" doesn\'t exist'
                    )
                else:
                    assert_never(instruction)
                method_diff = methods_we_should_have_applied_changes_to - methods_to_which_we_applied_changes
                if method_diff:
                    raise RouterGenerationError(
                        err.format(
                            endpoint_methods=list(method_diff),
                            endpoint_path=instruction.endpoint_path,
                            version_change_name=version_change.__name__,
                        ),
                    )


def _validate_no_repetitions_in_routes(routes: list[fastapi.routing.APIRoute]):
    route_map = {}

    for route in routes:
        route_info = _EndpointInfo(route.path, frozenset(route.methods))
        if route_info in route_map:
            raise RouteAlreadyExistsError(route, route_map[route_info])
        route_map[route_info] = route


def _add_data_migrations_to_route(
    route: APIRoute,
    head_route: Any,
    template_body_field: Union[type[BaseModel], None],
    template_body_field_name: Union[str, None],
    dependant_for_request_migrations: "Dependant",
    versions: VersionBundle,
):
    if not (route.dependant.request_param_name and route.dependant.response_param_name):  # pragma: no cover
        raise CadwynError(
            f"{route.dependant.request_param_name=}, {route.dependant.response_param_name=} "
            f"for route {list(route.methods)} {route.path} which should not be possible. Please, contact my author.",
        )

    route.endpoint = versions._versioned(
        template_body_field,
        template_body_field_name,
        route,
        head_route,
        dependant_for_request_migrations,
        request_param_name=route.dependant.request_param_name,
        background_tasks_param_name=route.dependant.background_tasks_param_name,
        response_param_name=route.dependant.response_param_name,
    )(route.endpoint)
    route.dependant.call = route.endpoint


def _apply_endpoint_had_instruction(
    version_change_name: str,
    instruction: EndpointHadInstruction,
    original_route: APIRoute,
    api_version_parameter_name: str,
    api_version_location: "APIVersionLocation",
):
    for attr_name in instruction.attributes.__dataclass_fields__:
        attr = getattr(instruction.attributes, attr_name)
        if attr is not Sentinel:
            if getattr(original_route, attr_name) == attr:
                raise RouterGenerationError(
                    f'Expected attribute "{attr_name}" of endpoint'
                    f' "{list(original_route.methods)} {original_route.path}"'
                    f' to be different in "{version_change_name}", but it was the same.'
                    " It means that your version change has no effect on the attribute"
                    " and can be removed.",
                )
            if attr_name == "path":
                original_path_params = {p.alias for p in original_route.dependant.path_params}
                new_path_params = set(re.findall("{(.*?)}", attr))
                if api_version_location == "path":
                    new_path_params.discard(api_version_parameter_name)
                if new_path_params != original_path_params:
                    raise RouterPathParamsModifiedError(
                        f'When altering the path of "{list(original_route.methods)} {original_route.path}" '
                        f'in "{version_change_name}", you have tried to change its path params '
                        f'from "{list(original_path_params)}" to "{list(new_path_params)}". It is not allowed to '
                        "change the path params of a route because the endpoint was created to handle the old path "
                        "params. In fact, there is no need to change them because the change of path params is "
                        "not a breaking change. If you really need to change the path params, you should create a "
                        "new route with the new path params and delete the old one.",
                    )
            setattr(original_route, attr_name, attr)


def _get_routes(
    routes: Sequence[BaseRoute],
    endpoint_path: str,
    endpoint_methods: set[str],
    endpoint_func_name: Union[str, None] = None,
    *,
    is_deleted: bool = False,
) -> list[fastapi.routing.APIRoute]:
    endpoint_path = endpoint_path.rstrip("/")
    return [
        route
        for route in routes
        if (
            isinstance(route, fastapi.routing.APIRoute)
            and route.path.rstrip("/") == endpoint_path
            and set(route.methods).issubset(endpoint_methods)
            and (endpoint_func_name is None or route.endpoint.__name__ == endpoint_func_name)
            and (_DELETED_ROUTE_TAG in route.tags) == is_deleted
        )
    ]


def _get_route_from_func(
    routes: Sequence[BaseRoute],
    endpoint: Endpoint,
) -> Union[fastapi.routing.APIRoute, None]:
    for route in routes:
        if isinstance(route, fastapi.routing.APIRoute) and (route.endpoint == endpoint):
            return route
    return None


def _route_has_a_simple_body_schema(route: APIRoute) -> bool:
    # Remember this: if len(body_params) == 1, then route.body_schema == route.dependant.body_params[0]
    return len(route.dependant.body_params) == 1



import re

import pytest
from pydantic import BaseModel

from cadwyn.exceptions import InvalidGenerationInstructionError
from cadwyn.structure.schemas import schema
from tests.conftest import CreateRuntimeSchemas, assert_models_are_equal, version_change


class MySchema(BaseModel):
    foo: str


def test__schema_had_name(create_runtime_schemas: CreateRuntimeSchemas):
    schemas = create_runtime_schemas(version_change(schema(MySchema).had(name="Aww")))

    assert_models_are_equal(schemas["2000-01-01"][MySchema], MySchema)
    assert schemas["2000-01-01"][MySchema].__name__ == "Aww"
    assert schemas["2000-01-01"][MySchema].__qualname__ == "Aww"


def test__schema_had_name__with_the_same_name__should_raise_error(create_runtime_schemas: CreateRuntimeSchemas):
    with pytest.raises(
        InvalidGenerationInstructionError,
        match=re.escape(
            'You tried to change the name of "MySchema" in "MyVersionChange" '
            "but it already has the name you tried to assign."
        ),
    ):
        create_runtime_schemas(version_change(schema(MySchema).had(name="MySchema")))
